"""
constants.py

This archive.py containing the constant or columns names used in some methods
and classes in the library

libraries:
    - pandera
"""
from pandera import Column, Check, DataFrameSchema

# Constants of DataFrames
DATE_COL = "START_DATETIME"
PRESSURE_COL = "PRESSURE_DATUM"
PRESSURE_PVT_COL = "Pressure"
TANK_COL = "Tank"
WELL_COL = "WELL_BORE"
OIL_CUM_COL = "OIL_CUM"
OIL_VOL_COL = "OIL_VOL"
WATER_CUM_COL = "WATER_CUM"
GAS_CUM_COL = "GAS_CUM"
LIQ_CUM = "LIQ_CUM"
OIL_FVF_COL = "Bo"
UO_COL = "uo"
GAS_FVF_COL = "Bg"
RS_COL = "GOR"
INJECTION_WATER = "WATER_VOL"
INFLUX_WATER = "We"
RS_W_COL = "RS_bw"
WATER_FVF_COL = "Bw"
UW_COL = "UW"
OIL_EXP = "Eo"
RES_EXP = "Efw"
OIL_CUM_TANK = "OIL_CUM_TANK"
WATER_CUM_TANK = "WATER_CUM_TANK"
GAS_CUM_TANK = "GAS_CUM_TANK"
WE = "Cumulative We"
OIL_RATE_COL = "OIL_RATE"
WATER_RATE_COL = "WATER_RATE"

# Frequency's of time
VALID_FREQS = ["D", "W", "M", "MS", "Q", "Y", None]

# VALIDATIONS:
_VECTOR_VALIDATION = Column(
    float,
    Check(lambda s: s >= 0),
    coerce=True,
    nullable=False,
)

_PRESSURE_VALIDATION = Column(float,
                              Check(lambda s: s >= 0),
                              coerce=True,
                              nullable=False,
                              required=False)

_STR_VALIDATION = Column(str,
                         Check(lambda s: s == s),
                         coerce=True,
                         nullable=False)

# SCHEMAS DICTIONARIES:
_PROD_SCHEMA_DICT = {
    OIL_CUM_COL: _VECTOR_VALIDATION,
    WATER_CUM_COL: _VECTOR_VALIDATION,
    GAS_CUM_COL: _VECTOR_VALIDATION,
    LIQ_CUM: _VECTOR_VALIDATION,
}

_PRESS_SCHEMA_DICT = {
    PRESSURE_COL: _PRESSURE_VALIDATION,
}

_INJ_SCHEMA_DICT = {INJECTION_WATER: _VECTOR_VALIDATION}

# SCHEMAS:
PROD_SCHEMA = DataFrameSchema(_PROD_SCHEMA_DICT, strict="filter")
PRESS_SCHEMA = DataFrameSchema(_PRESS_SCHEMA_DICT, strict="filter")
INJ_SCHEMA = DataFrameSchema(_INJ_SCHEMA_DICT, strict="filter")



"""
fluid_model.py

This archive.py is to calculate the PVT properties of oil and water using
linear interpolated.

libraries:
    - pydantic
    - scipy
    - pandera
"""

from pydantic import BaseModel
from scipy.interpolate import interp1d
import pandera as pa
from pandera.typing import DataFrame, Series
from pytank.constants.constants import (PRESSURE_PVT_COL,
                                        OIL_FVF_COL,
                                        GAS_FVF_COL,
                                        RS_COL)
from pytank.functions.pvt_correlations import RS_bw, Bo_bw


class _PVTSchema(pa.DataFrameModel):
    """
    Private Class to validate the values in the columns of PVT data
    """
    Pressure: Series[float] = pa.Field(ge=0,
                                       unique=True,
                                       coerce=True,
                                       nullable=False)
    Bo: Series[float] = pa.Field(ge=0, coerce=True)
    Bg: Series[float] = pa.Field(ge=0, coerce=True, nullable=True)
    GOR: Series[float] = pa.Field(ge=0, coerce=True)


class OilModel(BaseModel):
    """
    :parameter:
        - data_pvt: A DataFrame with the Oil PVT information that has
        been validated.
        - temperature (float): Temperature value [F].
    """
    data_pvt: DataFrame[_PVTSchema]
    temperature: float

    class Config:
        arbitrary_types_allowed = True

    def _interpolated_column_at_pressure(self, column_name: str,
                                         pressure: float) -> float:
        """

        :param:
            - column_name: Column name of PVT property of PVT DataFrame.
            - pressure: Pressure value to which the PVT property will be
            interpolated.

        :return:
            - float: Value of interpolated PVT property
        """
        df_pvt_local = self.data_pvt
        interp_func = interp1d(df_pvt_local[PRESSURE_PVT_COL],
                               df_pvt_local[column_name],
                               fill_value="extrapolate")
        return interp_func(pressure)

    def get_bo_at_press(self, pressure) -> float:
        """
        Method to interpolate oil volumetric factor Bo

        :param:
        pressure: Pressure value to which the PVT property will be
        interpolated.

        :return:
            - float: Value of Bo interpolated
        """
        return self._interpolated_column_at_pressure(OIL_FVF_COL, pressure)

    def get_bg_at_press(self, pressure) -> float:
        """
        Method to interpolate gas volumetric factor Bg

        :param:
        pressure: Pressure value to which the PVT property will be
        interpolated.

        :return:
            - float: Value of Bg interpolated
        """
        return self._interpolated_column_at_pressure(GAS_FVF_COL, pressure)

    def get_rs_at_press(self, pressure) -> float:
        """
        Method to interpolate oil solubility Rs

        :param:
        pressure: Pressure value to which the PVT property will be
        interpolated.

        :return:
            - float: Value of Rs interpolated
        """
        return self._interpolated_column_at_pressure(RS_COL, pressure)


class WaterModel(BaseModel):
    """

    :param:
        - salinity (float): Value of salinity [ppm]
        - temperature (float): Temperature value [F]
        - unit (int): 1 to Fields unit, 2 to Metric units

    """
    salinity: float = None
    temperature: float = None
    unit: int = None

    def get_bw_at_press(self, pressure: float) -> float:
        """
        Method to calculate the water volumetric factor Bw using correlation

        :param:
        pressure: Pressure value to calculate Bw

        :return:
            - float: Value of Bw
        """
        bw = Bo_bw(pressure, self.temperature, self.salinity, self.unit)
        return bw

    def get_rs_at_press(self, pressure: float) -> float:
        """
        Method to calculate the water solubility RS_bw using correlation

        :param:
        pressure: Pressure value to calculate RS_bw

        :return:
            - float: Value of RS_bw
        """
        rs = RS_bw(pressure, self.temperature, self.salinity, self.unit)
        return rs

    @staticmethod
    def get_default_bw() -> float:
        """
        Static method to calculate the Bw value per default.
        :return:
        float: 1
        """
        return float(1)

    @staticmethod
    def get_default_rs() -> float:
        """
        Static method to calculate the RS_bw value per default.
        :return:
        float: 0
        """
        return float(0)



"""
pvt_table.py

This module containing functions to calculate PVT properties If you do not
have a CSV of the PVT information.

Libraries:
    - pandas:
    - numpy

Important:
THE LOGIC OF THESE FUNCTIONS HAS NOT YET BEEN IMPLEMENTED TO THE LIBRARY.
SO ITS USE IS IN FUTURE DEVELOPMENT

"""
from pytank.functions import pvt_correlations as pvt
import pandas as pd
import numpy as np
import warnings

# Disable future warnings
warnings.simplefilter(action="ignore", category=FutureWarning)


def pvt_table1(p_sep,
               t_sep,
               api,
               rsp,
               sg_sep,
               tres,
               den_sto,
               p_res,
               salinity,
               jump,
               units=1) -> pd.DataFrame:
    """The pvt_table data frame is a table that has a size of the length of
    Pres(it is printed regarding a range) vs 6 columns, where within the
    first column are shown the pressure values (from 0 to Pres) whereas
    within each of the next 6 columns, are displayed 6 PC properties such as
    the oil density, FVF_oil, GOR, rsw, FVF_water, and water compressibility
    evaluated at the values of P from 0 to the reservoir pressure. The
    required input data is: Separator pressure [psia], Separator Temperature
    [F], Stock tank oil gravity [API], Separator producing GOR [scf/STB],
    Separator gas Specific Gravity, Reservoir Temperature [F], Density of
    stock-tank oil [lb/cu ft], Reservoir Pressure [psia], water salinity,
    and Jump(it indicates the step size of the iteration over the
    pressure).It also requires the Pb from Velarde correlation, which is
    called within the function.

    Parameters
    ----------
    p_sep: int or float
        Separator pressure [psia]
    t_sep: int or float
        Separator temperature [F]
    api: int or float
        Stock tank oil gravity [API]
    rsp: int or float
        Separator producing GOR [scf/STB]
    sg_sep: int or float
        Separator gas specific gravity
    tres: int or float
        Reservoir temperature [F]
    den_sto: int or float
        Density of stock-tank oil [lb/cu ft]
    p_res: int or float
        Initial reservoir pressure [psia]
    salinity: int or float
        Water Salinity [ppm]
    jump: int or float
        step size of the iteration over the pressure
    units: unit system
        if units = 1, it will use field units. Otherwise, it will be used
        metric units

    Returns
    -------
    Pandas Dataframe:
        Returns a pandas Dataframe with some pvt properties
    """

    # Call the bubble point pressure function
    pb = pvt.Pb_Velarde(api, sg_sep, tres, p_sep, t_sep, rsp, units)

    # Naming the dataframe columns
    pressure = "pressure[psia]"
    gor = "gor[scf/stb]"
    density = "density[lb/cu ft]"
    oil_fvf = "oil_fvf[rb/stb]"
    rsw = "gas_water_Rs[scf/stb]"
    water_fvf = "water_fvf[rb/stb]"
    water_comp = "water_comp[1/psia]"

    # List containing the names of all columns
    columns = [pressure, gor, density, oil_fvf, rsw, water_fvf, water_comp]

    # Creation of empty dataframes; under, at, and above the bubble
    # pressure point
    df_under = pd.DataFrame(columns=columns)
    df_pb = pd.DataFrame(columns=columns)
    df_above = pd.DataFrame(columns=columns)
    # creations of empty arrays
    array_under = []
    array_pb = []
    array_above = []

    # Loop to iterate from 0 to reservoir pressure
    for p in np.arange(0, p_res + 1, jump):

        # Condition when pressure is less than the bubble point pressure
        if 0 < p < pb:

            gor_velarde2 = pvt.Solution_GOR_Velarde2(sg_sep, api, tres, p,
                                                     p_sep, t_sep, rsp, units)
            den_underpb = pvt.Den_underPb(sg_sep, tres, p, api, p_sep, t_sep,
                                          rsp, units)
            fvf_underpb = pvt.FVF_underPb(den_sto, p_sep, t_sep, api, rsp,
                                          sg_sep, tres, p, units)
            rsw_under = pvt.RS_bw(p, tres, salinity, units)
            water_fvf_under = pvt.Bo_bw(p, tres, salinity, units)
            water_comp_under = pvt.comp_bw_nogas(p, tres, salinity, units)
            p_under = p

            array_under.append([
                p_under,
                gor_velarde2,
                den_underpb,
                fvf_underpb,
                rsw_under,
                water_fvf_under,
                water_comp_under,
            ])

        # Condition when pressures is equal to the bubble point pressure
        elif p == pb:
            array_pb.append([
                p,
                pvt.Solution_GOR_Pb_ValkoMcCain(p_sep, t_sep, api, rsp, units),
                pvt.Den_Pb(sg_sep, tres, p_sep, t_sep, api, rsp, units),
                pvt.FVF_Pb(den_sto, p_sep, t_sep, api, rsp, sg_sep, tres,
                           units),
                pvt.RS_bw(p, tres, salinity, units),
                pvt.Bo_bw(p, tres, salinity, units),
                pvt.comp_bw_nogas(p, tres, salinity, units),
            ])

        # Condition when pressure is greater than the bubble point pressure
        elif pb < p <= p_res:

            den_abovepb = pvt.Den_abovePb(p, sg_sep, tres, p_sep, t_sep, api,
                                          rsp, p_res, units)
            gor_velarde2_above = pvt.Solution_GOR_Pb_ValkoMcCain(
                p_sep, t_sep, api, rsp, units)
            fvf_abovepb = pvt.FVF_abovePb(p, den_sto, p_sep, t_sep, api, rsp,
                                          sg_sep, tres, p_res, units)
            rsw_above = pvt.RS_bw(p, tres, salinity, units)
            water_fvf_above = pvt.Bo_bw(p, tres, salinity, units)
            water_comp_above = pvt.comp_bw_nogas(p, tres, salinity, units)
            p_above = p

            array_above.append([
                p_above,
                gor_velarde2_above,
                den_abovepb,
                fvf_abovepb,
                rsw_above,
                water_fvf_above,
                water_comp_above,
            ])

    # Concatenation of the pvt dataframes regarding their relationship
    # to the bubble
    # Convert lists to numpy arrays
    array_under = np.array(array_under)
    array_pb = np.array(array_pb)
    array_above = np.array(array_above)

    # Create DataFrames
    df_above = pd.DataFrame(array_above)
    df_pb = pd.DataFrame(array_pb)
    df_under = pd.DataFrame(array_under)

    # Concatenate DataFrames
    pvt_dataframe = pd.concat([df_above, df_pb, df_under],
                              ignore_index=False,
                              axis=0)
    pvt_dataframe.columns = columns

    # Return the result
    return pvt_dataframe


def pvt_table2(p_sep,
               t_sep,
               api,
               rsp,
               sg_sep,
               tres,
               den_sto,
               p_res,
               salinity,
               jump,
               units=1) -> pd.DataFrame:
    """The pvt_table data frame is a table that has a size of the length of
    Pres(it is printed regarding a range) vs 6 columns, where within the
    first column are shown the pressure values (from 0 to Pres) whereas
    within each of the next 6 columns, are displayed 6 PC properties such as
    the oil density, FVF_oil, GOR, rsw, FVF_water, and water compressibility
    evaluated at the values of P from 0 to the reservoir pressure. The
    required input data is: Separator pressure [psia], Separator Temperature
    [F], Stock tank oil gravity [API], Separator producing GOR [scf/STB],
    Separator gas Specific Gravity, Reservoir Temperature [F], Density of
    stock-tank oil [lb/cu ft], Reservoir Pressure [psia], water salinity,
    and Jump(it indicates the step size of the iteration over the
    pressure).It also requires the Pb from Velarde correlation, which is
    called within the function.

    Parameters
    ----------
    p_sep: int or float
        Separator pressure [psia]
    t_sep: int or float
        Separator temperature [F]
    api: int or float
        Stock tank oil gravity [API]
    rsp: int or float
        Separator producing GOR [scf/STB]
    sg_sep: int or float
        Separator gas specific gravity
    tres: int or float
        Reservoir temperature [F]
    den_sto: int or float
        Density of stock-tank oil [lb/cu ft]
    p_res: int or float
        Initial reservoir pressure [psia]
    salinity: int or float
        Water Salinity [ppm]
    jump: int or float
        step size of the iteration over the pressure
    units: unit system
        if units = 1, it will use field units. Otherwise, it will be
        used metric units

    Returns
    -------
    Pandas Dataframe:
        Returns a pandas Dataframe with some pvt properties
    """

    # Call the bubble point pressure function
    pb = pvt.Pb_Velarde(api, sg_sep, tres, p_sep, t_sep, rsp, units)

    # Naming the dataframe columns
    pressure = "pressure[psia]"
    gor = "gor[scf/stb]"
    density = "density[lb/cu ft]"
    oil_fvf = "oil_fvf[rb/stb]"
    rsw = "gas_water_Rs[scf/stb]"
    water_fvf = "water_fvf[rb/stb]"
    water_comp = "water_comp[1/psia]"

    # List containing the names of all columns
    columns = [pressure, gor, density, oil_fvf, rsw, water_fvf, water_comp]

    # Creation of empty dataframes; under, at, and above the bubble
    # pressure point
    df_under = pd.DataFrame(columns=columns)
    df_pb = pd.DataFrame(columns=columns)
    df_above = pd.DataFrame(columns=columns)

    # Creation of empty list, under, at, and above the bubble pressure point
    data_under = []
    data_pb = []
    data_above = []

    # Loop to iterate from 0 to reservoir pressure
    for p in np.arange(0, p_res + 1, jump):

        # Condition when pressure is less than the bubble point pressure
        if (p > 0) and (p < pb):

            gor_velarde2 = pvt.Solution_GOR_Velarde2(sg_sep, api, tres, p,
                                                     p_sep, t_sep, rsp, units)
            den_underpb = pvt.Den_underPb(sg_sep, tres, p, api, p_sep, t_sep,
                                          rsp, units)
            fvf_underpb = pvt.FVF_underPb(den_sto, p_sep, t_sep, api, rsp,
                                          sg_sep, tres, p, units)
            rsw_under = pvt.RS_bw(p, tres, salinity, units)
            water_fvf_under = pvt.Bo_bw(p, tres, salinity, units)
            water_comp_under = pvt.comp_bw_nogas(p, tres, salinity, units)
            p_under = p

            data_under.append([
                p_under,
                gor_velarde2,
                den_underpb,
                fvf_underpb,
                rsw_under,
                water_fvf_under,
                water_comp_under,
            ])

        # Condition when pressures is equal to the bubble point pressure
        elif p == pb:
            data_pb.append([
                p,
                pvt.Solution_GOR_Pb_ValkoMcCain(p_sep, t_sep, api, rsp, units),
                pvt.Den_Pb(sg_sep, tres, p_sep, t_sep, api, rsp, units),
                pvt.FVF_Pb(den_sto, p_sep, t_sep, api, rsp, sg_sep, tres,
                           units),
                pvt.RS_bw(p, tres, salinity, units),
                pvt.Bo_bw(p, tres, salinity, units),
                pvt.comp_bw_nogas(p, tres, salinity, units),
            ])

        # Condition when pressure is greater than the bubble point pressure
        elif (p > pb) and (p <= p_res):

            den_abovepb = pvt.Den_abovePb(p, sg_sep, tres, p_sep, t_sep, api,
                                          rsp, p_res, units)
            gor_velarde2_above = pvt.Solution_GOR_Pb_ValkoMcCain(
                p_sep, t_sep, api, rsp, units)
            fvf_abovepb = pvt.FVF_abovePb(p, den_sto, p_sep, t_sep, api, rsp,
                                          sg_sep, tres, p_res, units)
            rsw_above = pvt.RS_bw(p, tres, salinity, units)
            water_fvf_above = pvt.Bo_bw(p, tres, salinity, units)
            water_comp_above = pvt.comp_bw_nogas(p, tres, salinity, units)
            p_above = p

            data_above.append([
                p_above,
                gor_velarde2_above,
                den_abovepb,
                fvf_abovepb,
                rsw_above,
                water_fvf_above,
                water_comp_above,
            ])
    # Empty dataframes
    df_under = pd.DataFrame(data_under, columns=columns)
    df_pb = pd.DataFrame(data_pb, columns=columns)
    df_above = pd.DataFrame(data_above, columns=columns)

    # Concatenation of the pvt dataframes regarding their relationship
    # to the bubble point pressure
    pvt_dataframe = pd.concat([df_under, df_pb, df_above],
                              ignore_index=True,
                              axis=0)

    return pvt_dataframe


# Example values for the function parameters
p_sep = 2500  # Separator pressure [psia]
t_sep = 150  # Separator temperature [F]
api = 35  # Stock tank oil gravity [API]
rsp = 500  # Separator producing GOR [scf/STB]
sg_sep = 0.7  # Separator gas specific gravity
tres = 180  # Reservoir temperature [F]
den_sto = 55  # Density of stock-tank oil [lb/cu ft]
p_res = 5000  # Initial reservoir pressure [psia]
salinity = 500  # Water Salinity [ppm]
jump = 100  # step size of the iteration over the pressure

# Call the function with these example values
result_df = pvt_table2(p_sep, t_sep, api, rsp, sg_sep, tres, den_sto, p_res,
                       salinity, jump)



"""
vector_data.py

This module defines the VectorData Class for handling vector data with
specific validations schemas and date indices.

The logic is structured using classes and methods.

libraries:
    - pandas
    - datetime
    - typing
    - pydantic
    - pandera
"""

import pandas as pd
from datetime import datetime
from typing import Any
from pydantic import BaseModel, PrivateAttr, validator
from pandera import DataFrameSchema
import pandera as pa
from pytank.functions.utilities import add_date_index_validation
from pytank.constants.constants import (
    PROD_SCHEMA,
    PRESS_SCHEMA,
    INJ_SCHEMA,
)


class VectorData(BaseModel):
    """
    Class used to handle vector data with a specific validation scheme
    and date index.

    :parameter:
        - is_result: Indicates whether the vector is a result.
        - data_schema: Validation scheme for the data.
        - freq (str): Frequency of the data.
        - use_pressure: Indicates whether the pressure is used.
        - data: Data container.
        - _start_date and _end_date: Private start and end dates.
    """
    is_result: bool = False
    data_schema: DataFrameSchema = DataFrameSchema()
    freq: str | None
    use_pressure: bool = False
    data: Any
    _start_date: datetime = PrivateAttr(None)
    _end_date: datetime = PrivateAttr(None)

    class Config:
        arbitrary_types_allowed = True

    @validator("data")
    def validate_data(cls, v, values):
        """
        Validate the dates with the specific scheme
        :param
            - v:
            - values:
        :return:
            - data_schema
        """
        new_schema = add_date_index_validation(values["data_schema"],
                                               values["freq"])

        cls.data_schema = new_schema
        return new_schema.validate(v)

    @property
    def start_date(self):
        """
        :return:
            - start date
        """
        if self._start_date is None:
            self._start_date = self.data.index.min()
        return self._start_date

    @property
    def end_date(self):
        """
        :return:
            - end date
        """
        if self._end_date is None:
            self._end_date = self.data.index.max()
        return self._end_date

    def equal_date_index(self, other) -> bool:
        """
        Compare the vector_data objects

        :param other:

        :return:
        A comparative
        """
        return all([
            self.start_date == other.start_date,
            self.end_date == other.end_date,
            self.freq == other.freq,
        ])

    def get_date_index(self) -> pd.DatetimeIndex:
        """
        :return:
            - Date index of data
        """
        return self.data.index

    def _eq_(self, other):
        """
        Compare two VectorDara Objects
        """
        return all([
            self.data_schema == other.data_schema,
            self.start_date == other.start_date,
            self.end_date == other.end_date,
        ])

    def _len_(self):
        """
        :return:
            - The length of the data
        """
        return len(self.data)

    def _add_(self, other):
        """
        Allows the addition of two VectorData objects or a VectorData with a
        number or a series.
        """
        if isinstance(other, VectorData):
            if self == other:
                # If the two VectorData have the same schema, then we
                # can just add them together using a groupby sum on
                # the date index
                new_data = pd.concat([self.data,
                                      other.data]).groupby(level=0).sum()
                return VectorData(
                    data_schema=self.data_schema,
                    freq=self.freq,
                    # use_pressure=self.use_pressure,
                    data=new_data,
                )
            elif self.equal_date_index(other):
                # If the two VectorData have the same date index, but different
                # schemas, then we need to add them together using a concat
                # on thecolumns that are in neither dataframe and a groupby
                # sum on the columnsthat are in both dataframes
                common_cols = self.data.columns.intersection(
                    other.data.columns)
                left_cols = self.data.columns.difference(other.data.columns)
                right_cols = other.data.columns.difference(self.data.columns)
                new_data_common = pd.DataFrame()
                new_data_left = pd.DataFrame()
                new_data_right = pd.DataFrame()
                if len(common_cols) > 0:
                    new_data_common = (pd.concat(
                        [self.data[common_cols],
                         other.data[common_cols]]).groupby(level=0).sum())
                if len(left_cols) > 0:
                    new_data_left = self.data[left_cols]
                if len(right_cols) > 0:
                    new_data_right = other.data[right_cols]

                new_data = pd.concat(
                    [new_data_common, new_data_left, new_data_right],
                    axis=1)
                return VectorData(
                    data_schema=pa.infer_schema(new_data),
                    freq=self.freq,
                    # use_pressure=self.use_pressure,
                    data=new_data,
                )
            else:
                raise ValueError(
                    "The date index of the two VectorData objects are not "
                    "equal"
                )
        elif isinstance(other, (int, float)):
            new_data = self.data + other
            return VectorData(
                data_schema=self.data_schema,
                freq=self.freq,
                # use_pressure=self.use_pressure,
                data=new_data,
            )
        elif isinstance(other, pd.Series):
            if len(self) == len(other):
                new_data = self.data + other
                return VectorData(
                    data_schema=self.data_schema,
                    freq=self.freq,
                    # use_pressure=self.use_pressure,
                    data=new_data,
                )

    def _radd_(self, other):
        """
        Reverse addition to allow addition of a number with a VectorData.
        """
        return self.add(other)


class ProdVector(VectorData):
    """
    This class is used to handle production data with a specific scheme.
    """
    data_schema: DataFrameSchema = PROD_SCHEMA


class PressVector(VectorData):
    """
    This class is used to handle pressure data with a specific scheme.
    """
    data_schema: DataFrameSchema = PRESS_SCHEMA


class InjVector(VectorData):
    """
    This class is used to handle injection data with a specific scheme.
    """
    data_schema: DataFrameSchema = INJ_SCHEMA



"""
tank.py

This module defines the Tank CLass is to process data internally.

Also, it works as a container for the reservoir properties.

libraries:
    - pandas
    - pydantic
    - typing
"""

import pandas as pd
from pydantic import BaseModel
from typing import Union, Optional
from pytank.constants.constants import (
    OIL_FVF_COL,
    GAS_FVF_COL,
    RS_COL,
    PRESSURE_COL,
    OIL_CUM_COL,
    GAS_CUM_COL,
    WATER_CUM_COL,
    DATE_COL,
    WELL_COL,
    WATER_FVF_COL,
    RS_W_COL,
    TANK_COL,
    LIQ_CUM,
)
from pytank.fluid_model import OilModel, WaterModel
from pytank.aquifer_model import Fetkovich, CarterTracy


class Tank(BaseModel):
    """
    Class that functions as a container for the reservoir (tank) properties.
    """
    name: str
    wells: list
    oil_model: OilModel
    water_model: WaterModel
    pi: float
    swo: float
    cw: float
    cf: float
    aquifer: Optional[Union[None, Fetkovich, CarterTracy]]

    class Config:
        arbitrary_types_allowed = True

    def __init__(self, name: str, wells: list, oil_model: OilModel,
                 water_model: WaterModel, pi: float, swo: float, cw: float,
                 cf: float, aquifer: Optional[Union[None, Fetkovich,
                                                    CarterTracy]]):
        """
        Init Method
        :param
            - name: Name of tank (reservoir)
            - wells: Instance of Well Class (list)
            - oil_model: Instance of OilModel Class
            - water_model: Instance of WaterModel CLass
            - pi: Initial Pressure [Psi]
            - swo: Water initial saturation [decimal]
            - cw: Water compressibility
            - cf: Total compressibility
            - aquifer: Instance of Aquifer Class
        """
        super().__init__(name=name,
                         wells=wells,
                         oil_model=oil_model,
                         water_model=water_model,
                         pi=pi,
                         swo=swo,
                         cw=cw,
                         cf=cf,
                         aquifer=aquifer)

    def _press_df_internal(self) -> pd.DataFrame:
        """
        Private method that internally manages the pressure vector for use in
        the UW method.

        :return:
            - pd.Dataframe: A pressure DataFrame with properties PVT of oil
             and water
        """
        df_press = pd.DataFrame()
        for well in self.wells:
            press_vector = well.press_data
            well_name = well.name
            tank_name = self.name
            if press_vector is not None:

                well_date = press_vector.data.index
                well_oil_fvf = self.oil_model.get_bo_at_press(
                    press_vector.data[PRESSURE_COL])
                well_gas_fvf = self.oil_model.get_bg_at_press(
                    press_vector.data[PRESSURE_COL])
                well_rs = self.oil_model.get_rs_at_press(
                    press_vector.data[PRESSURE_COL])

                # In case properties are calculated using correlations
                if (self.water_model.salinity is not None
                        and self.water_model.temperature is not None
                        and self.water_model.unit is not None):
                    well_bw = self.water_model.get_bw_at_press(
                        press_vector.data[PRESSURE_COL])
                    well_rs_w = self.water_model.get_rs_at_press(
                        press_vector.data[PRESSURE_COL])

                    # In case there are default values for Bw and Rs_w
                else:
                    well_bw = self.water_model.get_default_bw()
                    well_rs_w = self.water_model.get_default_rs()

                # Create a copy of data from press_vector
                temp_df_press = press_vector.data.copy()

                # Add columns to DataFrame
                temp_df_press[WELL_COL] = well_name
                temp_df_press[DATE_COL] = well_date
                temp_df_press[OIL_FVF_COL] = well_oil_fvf
                temp_df_press[GAS_FVF_COL] = well_gas_fvf
                temp_df_press[RS_COL] = well_rs
                temp_df_press[WATER_FVF_COL] = well_bw
                temp_df_press[RS_W_COL] = well_rs_w
                temp_df_press[TANK_COL] = tank_name

                df_press = pd.concat([df_press, temp_df_press],
                                     ignore_index=True)
        return df_press

    def get_pressure_df(self) -> pd.DataFrame:
        """
        Gets a DatFrame with pressure data using the private _press_df_internal
        method
        THIS METHOD IS MARKED PRIVATE FOR INTERNAL USE ONLY.

        :return:
            pd.Dataframe: DataFrame with PVT properties.
        """
        return self._press_df_internal()

    def _prod_df_internal(self) -> pd.DataFrame:
        """
        Private method that internally manages production vector for use
        in the UW method.

        :return:
            - pd.Dataframe: A production DataFrame

        """
        df_prod = pd.DataFrame()
        for well in self.wells:
            prod_vector = well.prod_data
            well_name = well.name
            tank_name = self.name
            if prod_vector is not None:
                well_date = prod_vector.data.index
                well_oil_cum = prod_vector.data[OIL_CUM_COL]
                well_water_cum = prod_vector.data[WATER_CUM_COL]
                well_gas_cum = prod_vector.data[GAS_CUM_COL]
                well_liq_cum = prod_vector.data[LIQ_CUM]

                # Create a copy of data from prod_vector
                temp_df_prod = prod_vector.data.copy()

                temp_df_prod[WELL_COL] = well_name
                temp_df_prod[DATE_COL] = well_date
                temp_df_prod[OIL_CUM_COL] = well_oil_cum
                temp_df_prod[WATER_CUM_COL] = well_water_cum
                temp_df_prod[GAS_CUM_COL] = well_gas_cum
                temp_df_prod[LIQ_CUM] = well_liq_cum
                temp_df_prod[TANK_COL] = tank_name

                df_prod = pd.concat([df_prod, temp_df_prod],
                                    ignore_index=True)
        return df_prod

    def get_production_df(self):
        """
        Gets a DataFrame with production data using the private
        _prod_df_internal method
        THIS METHOD IS MARKED PRIVATE FOR INTERNAL USE ONLY.

        :return:
            pd.DataFrame: A DataFrame with cumulative production
        """
        return self._prod_df_internal()



import pandas as pd
from typing import Optional, List
from pytank.constants.constants import (OIL_CUM_COL, WATER_CUM_COL,
                                        GAS_CUM_COL, LIQ_CUM, PRESSURE_COL,
                                        DATE_COL)
from pytank.vector_data import ProdVector, PressVector
from pandera.errors import SchemaError
from pytank.functions.utilities import normalize_date_freq
import warnings
from pytank.well import Well


def create_wells(df_prod: pd.DataFrame,
                 df_press: pd.DataFrame,
                 freq_prod: Optional[str] = None,
                 freq_press: Optional[str] = None) -> List[Well]:
    """
    Parameters
    ----------
    df_prod : pandas.DataFrame
        pd.DataFrame containing the production data.
    df_press : pandas.DataFrame
        pd.DataFrame containing the pressure data.
    freq_prod : str
        Frequency of the production data. Can be None if the frequency
        is correct.
    freq_press : str
        Frequency of the pressure data. It is not necessary.

    Returns
        -------
        List[Wells]
            A list of `Wells` objects, where each object represents a well
            with its associated pressure and production data.
    """
    warnings.filterwarnings(
        "ignore", message="DataFrame.fillna with 'method' "
                          "is deprecated")

    def _process_data(df_prod: pd.DataFrame, df_press: pd.DataFrame):
        """
        PRIVATE internal method to handle the production and pressure
        data (dates)
        :return:
            - Production DataFrame
            - Pressure DataFrame
        """
        prod_data = df_prod
        prod_data[DATE_COL] = pd.to_datetime(prod_data[DATE_COL])
        prod_data.set_index(prod_data[DATE_COL], inplace=True)

        press_data = df_press
        press_data[DATE_COL] = pd.to_datetime(press_data["DATE"])
        press_data = press_data.drop("DATE", axis=1)

        return prod_data, press_data

    prod_data, press_data = _process_data(df_prod, df_press)
    cols_fills_na = [OIL_CUM_COL, WATER_CUM_COL, GAS_CUM_COL, LIQ_CUM]
    all_wells = set(prod_data["ITEM_NAME"]).union(press_data["WELLBORE"])
    list_wells = []

    for name in all_wells:
        prod_vector = None
        press_vector = None

        if name in prod_data["ITEM_NAME"].unique():
            group_prod = prod_data[prod_data["ITEM_NAME"] == name]

            group_prod = group_prod.rename(
                columns={
                    OIL_CUM_COL: OIL_CUM_COL,
                    WATER_CUM_COL: WATER_CUM_COL,
                    GAS_CUM_COL: GAS_CUM_COL,
                })
            group_prod[LIQ_CUM] = group_prod[OIL_CUM_COL] + group_prod[
                WATER_CUM_COL]
            group_prod = group_prod[[
                OIL_CUM_COL, WATER_CUM_COL, GAS_CUM_COL, LIQ_CUM
            ]]

            # Normalize the frequency
            if freq_prod is not None:
                group_prod_norm = normalize_date_freq(
                    df=group_prod,
                    freq=freq_prod,
                    cols_fill_na=cols_fills_na,
                    method_no_cols="ffill")
                try:
                    prod_vector = ProdVector(freq=freq_prod,
                                             data=group_prod_norm)
                except SchemaError as e:
                    expected_error_msg = \
                        ('ValueError("Need at least 3 dates to '
                         'infer frequency")')
                    if str(e) == expected_error_msg:
                        group_prod.index.freq = freq_prod

                        # Create a production vector
                        prod_vector = ProdVector(freq=None,
                                                 data=group_prod_norm)

            else:
                prod_vector = ProdVector(freq=freq_prod,
                                         data=group_prod)

        if name in press_data["WELLBORE"].unique():
            group_press = press_data[press_data["WELLBORE"] == name]

            group_press = group_press.rename(columns={
                PRESSURE_COL: PRESSURE_COL,
            })
            group_press.set_index(DATE_COL, inplace=True)

            # Create a pressure vector
            press_vector = PressVector(freq=freq_press,
                                       data=group_press)

        # Create well lists
        info_well = Well(name=name,
                         prod_data=prod_vector,
                         press_data=press_vector)

        # Add wells list to tanks dict
        list_wells.append(info_well)

    return list_wells


def search_wells(wells: List[Well], well_names: List[str]) -> List[Well]:
    """
    Searches for wells in the list of all wells based on the provided well
    names.

    Parameters
    ----------
    wells: List of all wells
    well_names: List of well names to search for.

    Returns
    -------
    List[Wells]
        A list of `Wells` objects that match the provided well names.
    """
    result = [well for well in wells if well.name in well_names]

    # Well no found
    found_well_names = [well.name for well in result]
    not_found_wells = [name for name in well_names if name
                       not in found_well_names]
    # Warning
    if not_found_wells:
        warnings.warn(f"The following wells were not found in the list: "
                      f"{', '.join(not_found_wells)}")

    return result



"""
well.py

This module defines the Well Class to group production and pressure
information per well.

The logic is structured using classes and methods.

"""
from pydantic import BaseModel
from typing import Optional
from pytank.vector_data import ProdVector, PressVector


class Well(BaseModel):
    """
    Class used to handle pressure and production vectors
    """
    name: str
    prod_data: Optional[ProdVector] = None
    press_data: Optional[PressVector] = None



"""
pvt_correlations.py

This module containing a lot of functions to calculate PVT properties using
 correlations

libraries:
    - numpy

Importance:
NOT ALL THE FUNCTIONS OF THIS MODULE ARE USED. HOWEVER, THEY CAN BE USED
IN FUTURE UPDATES OF THE LIBRARY
"""
import numpy as np
"----------------------------Oil Correlations-----------------------------"


def Solution_GOR_Pb_ValkoMcCain(Psep, Tsep, API, Rsp, Units=1):
    """The ValkoMcCain correlation calculates the solution GOR at
    bubblepoint pressure, the required input data is: Separator pressure [
    psia], Separator Temperature [F], Separator producing GOR [scf/STB] and
    Stock tank oil gravity [API]

    If Units=1 use:
        Pressure in psia
        Temperature in F
        Stock-Tank oil gravity in API
        Solution Gas Oil Ratio @ Pb in Scf/STB

    If Units=2 use:
        Pressure in MPa
        Temperature in C
        Stock-Tank oil gravity in API
        Solution Gas Oil Ratio @ Pb in m3/m3"""

    if Units == 1:

        Psep_array = np.array(Psep)
        Tsep_array = np.array(Tsep)
        API_array = np.array(API)
        Rsp_array = np.array(Rsp)

    else:
        Psep_array = np.array(Psep)
        Psep_array = Psep_array / (1.45037738 * 10**2)
        Tsep_array = np.array(Tsep)
        Tsep_array = (Tsep_array * 9 / 5) + 32
        API_array = np.array(API)
        Rsp_array = np.array(Rsp)
        Rsp_array = Rsp_array * (5.61458333)

    if Psep_array > 0 and Tsep_array > 0 and API_array > 0 and Rsp_array > 0:

        lnPsep = np.log(Psep_array)
        lnTsep = np.log(Tsep_array)

        #              VARn                C0n            C1n             C2n
        Zntable = {
            "VAR1": lnPsep,
            "C01": -8.005,
            "C11": 2.7,
            "C21": -0.161,
            "VAR2": lnTsep,
            "C02": 1.224,
            "C12": -0.5,
            "C22": 0,
            "VAR3": API_array,
            "C03": -1.587,
            "C13": 0.0441,
            "C23": -2.29e-5
        }

        z1 = Zntable["C01"] + (Zntable["C11"] * Zntable["VAR1"]) + (
            Zntable["C21"] * Zntable["VAR1"]**2)
        z2 = Zntable["C02"] + (Zntable["C12"] * Zntable["VAR2"]) + (
            Zntable["C22"] * Zntable["VAR2"]**2)
        z3 = Zntable["C03"] + (Zntable["C13"] * Zntable["VAR3"]) + (
            Zntable["C23"] * Zntable["VAR3"]**2)

        sumZ = z1 + z2 + z3
        Rst = np.exp(3.955 + 0.83 * sumZ - 0.024 * (sumZ**2) + (0.075 *
                                                                (sumZ**3)))
        Rsb = Rst + Rsp

    elif (Psep_array == 0 or Tsep_array == 0 or API_array ==
          0 and Rsp_array > 0):

        Rsb = 1.1618 * Rsp_array

    else:
        Rsb = "Incorrect Data"

    if Units == 1:

        Rsb = Rsb

    else:
        Rsb = Rsb / (1.78107607 * 10**-1)

    return Rsb


def Spec_grav_st_ValkoMcCain2(Psep, Tsep, API, Rsp, SGsep, Units=1):
    """ The Spec_grav_st_ValkoMcCain2 correlation calculates the
    weighted-average Specific Gravities of surface gases at Stock-Tank
    conditions, the required imput data is: Separator pressure [psia],
    Separator Temperature [F], Separator producing GOR [scf/STB], Separator
    gas Specific Gravity and Stock tank oil gravity [API]

        If Units=1 use:
            Pressure in psia
            Temperature in F
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in Scf/STB

        If Units=2 use:
            Pressure in MPa
            Temperature in C
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in m3/m3"""

    if Units == 1:

        Psep_array = np.array(Psep)
        Tsep_array = np.array(Tsep)
        API_array = np.array(API)
        Rsp_array = np.array(Rsp)
        SGsep_array = np.array(SGsep)

    else:
        Psep_array = np.array(Psep)
        Psep_array = Psep_array / (1.45037738 * 10**2)
        Tsep_array = np.array(Tsep)
        Tsep_array = (Tsep_array * 9 / 5) + 32
        API_array = np.array(API)
        Rsp_array = np.array(Rsp)
        Rsp_array = Rsp_array * (5.61458333)
        SGsep_array = np.array(SGsep)

    if (Psep_array > 0 and Tsep_array > 0 and API_array >
            0 and Rsp_array > 0 and SGsep_array > 0):

        lnPsep = np.log(Psep_array)
        lnRsp = np.log(Rsp_array)

        # VARn            C0n            C1n             C2n    C3n
        # C4n
        Zntable = {
            "VAR1": lnPsep,
            "C01": -17.275,
            "C11": 7.9597,
            "C21": -1.1013,
            "C31": 2.7735e-2,
            "C41": 3.2287e-3,
            "VAR2": lnRsp,
            "C02": -0.3354,
            "C12": -0.3346,
            "C22": 0.1956,
            "C32": -3.4374e-2,
            "C42": 2.08e-3,
            "VAR3": API_array,
            "C03": 3.705,
            "C13": -0.4273,
            "C23": 1.818e-2,
            "C33": -3.459e-4,
            "C43": 2.505e-6,
            "VAR4": SGsep_array,
            "C04": -155.52,
            "C14": 629.61,
            "C24": -957.38,
            "C34": 647.57,
            "C44": -163.26,
            "VAR5": Tsep_array,
            "C05": 2.085,
            "C15": -7.097e-2,
            "C25": 9.859e-4,
            "C35": -6.312e-6,
            "C45": 1.4e-8
        }

        z1 = Zntable["C01"] + (Zntable["C11"] * Zntable["VAR1"]) + (
            Zntable["C21"] * Zntable["VAR1"]**2) + (
                Zntable["C31"] * Zntable["VAR1"]**3) + (Zntable["C41"] *
                                                        Zntable["VAR1"]**4)
        z2 = Zntable["C02"] + (Zntable["C12"] * Zntable["VAR2"]) + (
            Zntable["C22"] * Zntable["VAR2"]**2) + (
                Zntable["C32"] * Zntable["VAR2"]**3) + (Zntable["C42"] *
                                                        Zntable["VAR2"]**4)
        z3 = Zntable["C03"] + (Zntable["C13"] * Zntable["VAR3"]) + (
            Zntable["C23"] * Zntable["VAR3"]**2) + (
                Zntable["C33"] * Zntable["VAR3"]**3) + (Zntable["C43"] *
                                                        Zntable["VAR3"]**4)
        z4 = Zntable["C04"] + (Zntable["C14"] * Zntable["VAR4"]) + (
            Zntable["C24"] * Zntable["VAR4"]**2) + (
                Zntable["C34"] * Zntable["VAR4"]**3) + (Zntable["C44"] *
                                                        Zntable["VAR4"]**4)
        z5 = Zntable["C05"] + (Zntable["C15"] * Zntable["VAR5"]) + (
            Zntable["C25"] * Zntable["VAR5"]**2) + (
                Zntable["C35"] * Zntable["VAR5"]**3) + (Zntable["C45"] *
                                                        Zntable["VAR5"]**4)
        sumZ = z1 + z2 + z3 + z4 + z5

        SGst = 1.219 + 0.198 * sumZ + 0.0845 * (sumZ**2) + (
            0.03 * (sumZ**3)) + (0.003 * (sumZ**4))

    else:
        SGst = "Incorrect Data"

    return SGst


# print("SGst =",Spec_grav_st_ValkoMcCain2(Psep, Tsep, API, Rsp, SGsep))


def Spec_grav_ValkoMcCain2(Psep, Tsep, API, Rsp, SGsep, Units=1):
    """ The ValkoMcCain2 correlation calculates the weighted-average
    Specific Gravities of surface gases, the required imput data is:
    Separator pressure [psia], Separator Temperature [F], Separator
    producing GOR [scf/STB], Separator gas Specific Gravity and Stock tank
    oil gravity [API]

        If Units=1 use:
            Pressure in psia
            Temperature in F
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in Scf/STB

        If Units=2 use:
            Pressure in MPa
            Temperature in C
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in m3/m3"""

    if Units == 1:

        Psep_array = np.array(Psep)
        Tsep_array = np.array(Tsep)
        API_array = np.array(API)
        Rsp_array = np.array(Rsp)
        SGsep_array = np.array(SGsep)

    else:

        Psep_array = np.array(Psep)
        Psep_array = Psep_array / (1.45037738 * 10**2)
        Tsep_array = np.array(Tsep)
        Tsep_array = (Tsep_array * 9 / 5) + 32
        API_array = np.array(API)
        Rsp_array = np.array(Rsp)
        Rsp_array = Rsp_array * (5.61458333)
        SGsep_array = np.array(SGsep)

    if (Psep_array > 0 and Tsep_array > 0 and API_array >
            0 and Rsp_array > 0 and SGsep_array > 0):

        lnPsep = np.log(Psep_array)
        lnTsep = np.log(Tsep_array)

        #              VARn                C0n            C1n             C2n
        Zntable = {
            "VAR1": lnPsep,
            "C01": -8.005,
            "C11": 2.7,
            "C21": -0.161,
            "VAR2": lnTsep,
            "C02": 1.224,
            "C12": -0.5,
            "C22": 0,
            "VAR3": API_array,
            "C03": -1.587,
            "C13": 0.0441,
            "C23": -2.29e-5
        }

        z1 = Zntable["C01"] + (Zntable["C11"] * Zntable["VAR1"]) + (
            Zntable["C21"] * Zntable["VAR1"]**2)
        z2 = Zntable["C02"] + (Zntable["C12"] * Zntable["VAR2"]) + (
            Zntable["C22"] * Zntable["VAR2"]**2)
        z3 = Zntable["C03"] + (Zntable["C13"] * Zntable["VAR3"]) + (
            Zntable["C23"] * Zntable["VAR3"]**2)

        sumZ = z1 + z2 + z3
        Rst = np.exp(3.955 + 0.83 * sumZ - 0.024 * (sumZ**2) + (0.075 *
                                                                (sumZ**3)))

        SG = ((SGsep_array * Rsp_array) +
              (Spec_grav_st_ValkoMcCain2(Psep, Tsep, API, Rsp, SGsep) *
               Rst)) / (Rsp_array + Rst)

    elif (Psep_array == 0 or Tsep_array == 0 or API_array ==
          0 or Rsp_array == 0 and SGsep_array > 0):

        SG = 1.066 * SGsep_array

    else:
        SG = "Incorrect Data"

    return SG


# print("SG =",Spec_grav_ValkoMcCain2 (Psep, Tsep, API, Rsp, SGsep))


def Pb_Velarde(API, SGsep, Tres, Psep, Tsep, Rsp, Units=1):
    """The Velarde correlation calculates the Bubblepoint Pressures at
    Reservoir Temperatures, the required imput data is: Reservoir
    Temperature [F], Separator gas Specific Gravity and Stock tank oil
    gravity [API]. Also requires the Rsb from the ValkoMcCain correlation

        If Units=1 use:
            Pressure in psia
            Temperature in F
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in Scf/STB

        If Units=2 use:
            Pressure in MPa
            Temperature in C
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in m3/m3"""

    if Units == 1:

        API_array = np.array(API)
        SGsep_array = np.array(SGsep)
        Tres_array = np.array(Tres)

    else:

        Tres_array = np.array(Tres)
        Tres_array = (Tres_array * 9 / 5) + 32
        API_array = np.array(API)
        SGsep_array = np.array(SGsep)

    if API_array > 0 and SGsep_array > 0 and Tres_array > 0:

        lnRsb = np.log(Solution_GOR_Pb_ValkoMcCain(Psep, Tsep, API, Rsp))

        # VARn                   C0n              C1n                C2n
        # C3n
        Zntable = {
            "VAR1": lnRsb,
            "C01": -5.48,
            "C11": -0.0378,
            "C21": 0.281,
            "C31": -0.0206,
            "VAR2": API_array,
            "C02": 1.27,
            "C12": -0.0449,
            "C22": 4.36e-4,
            "C32": -4.76e-6,
            "VAR3": SGsep_array,
            "C03": 4.51,
            "C13": -10.84,
            "C23": 8.39,
            "C33": -2.34,
            "VAR4": Tres_array,
            "C04": -0.7835,
            "C14": 6.23e-3,
            "C24": -1.22e-5,
            "C34": 1.03e-8
        }

        z1 = Zntable["C01"] + (Zntable["C11"] * Zntable["VAR1"]) + (
            Zntable["C21"] * Zntable["VAR1"]**2) + (Zntable["C31"] *
                                                    Zntable["VAR1"]**3)
        z2 = Zntable["C02"] + (Zntable["C12"] * Zntable["VAR2"]) + (
            Zntable["C22"] * Zntable["VAR2"]**2) + (Zntable["C32"] *
                                                    Zntable["VAR2"]**3)
        z3 = Zntable["C03"] + (Zntable["C13"] * Zntable["VAR3"]) + (
            Zntable["C23"] * Zntable["VAR3"]**2) + (Zntable["C33"] *
                                                    Zntable["VAR3"]**3)
        z4 = Zntable["C04"] + (Zntable["C14"] * Zntable["VAR4"]) + (
            Zntable["C24"] * Zntable["VAR4"]**2) + (Zntable["C34"] *
                                                    Zntable["VAR4"]**3)
        sumZ = z1 + z2 + z3 + z4

        Pb = np.exp(7.475 + 0.713 * sumZ + 0.0075 * (sumZ**2))

    else:
        Pb = "Missing Data"

    if Units == 1:
        Pb = Pb

    else:
        Pb = Pb * (6.89475729 * 10**-3)

    return Pb


# print ("Pb =",Pb_Velarde (API, SGsep, Tres, Psep, Tsep, Rsp))


def Solution_GOR_Velarde2(SGsep, API, Tres, P, Psep, Tsep, Rsp, Units=1):
    """The Velarde2 correlation calculates the Solution Gas-Oil ratios at
    reservoir reservoir pressures less than bubblepoint pressure,
    the required imput data is: Separator gas Specific Gravity, Stock tank
    oil gravity [API], Reservoir Temperature [F] and Pressure [psia]. Also
    requires the Pb from the Velarde correlation and Rsb from the
    ValkoMcCain correlation

        If Units=1 use:
            Pressure in psia
            Temperature in F
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in Scf/STB

        If Units=2 use:
            Pressure in MPa
            Temperature in C
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in m3/m3"""

    if Units == 1:

        SGsep_array = np.array(SGsep)
        API_array = np.array(API)
        Tres_array = np.array(Tres)
        P_array = np.array(P)

    else:
        Tres_array = np.array(Tres)
        Tres_array = (Tres_array * 9 / 5) + 32
        API_array = np.array(API)
        SGsep_array = np.array(SGsep)
        P_array = np.array(P)
        P_array = P_array / (1.45037738 * 10**2)

    if (API_array > 0 and SGsep_array > 0 and Tres_array > 0 and P_array <
            Pb_Velarde(
            API, SGsep, Tres, Psep, Tsep, Rsp)):

        Pr = (P_array -
              14.7) / (Pb_Velarde(API, SGsep, Tres, Psep, Tsep, Rsp) - 14.7)

        #              An                Bn               Cn
        antable = {
            "A0": 9.73e-7,
            "B0": 0.022339,
            "C0": 0.725167,
            "A1": 1.672608,
            "B1": -1.00475,
            "C1": -1.485480,
            "A2": 0.92987,
            "B2": 0.337711,
            "C2": -0.164741,
            "A3": 0.247235,
            "B3": 0.132795,
            "C3": -0.09133,
            "A4": 1.056052,
            "B4": 0.302065,
            "C4": 0.047094
        }

        a1 = antable["A0"] * SGsep_array**antable["A1"] * API_array**antable[
            "A2"] * Tres_array**antable["A3"] * (Pb_Velarde(
                API, SGsep, Tres, Psep, Tsep, Rsp) - 14.7)**antable["A4"]
        a2 = antable["B0"] * SGsep_array**antable["B1"] * API_array**antable[
            "B2"] * Tres_array**antable["B3"] * (Pb_Velarde(
                API, SGsep, Tres, Psep, Tsep, Rsp) - 14.7)**antable["B4"]
        a3 = antable["C0"] * SGsep_array**antable["C1"] * API_array**antable[
            "C2"] * Tres_array**antable["C3"] * (Pb_Velarde(
                API, SGsep, Tres, Psep, Tsep, Rsp) - 14.7)**antable["C4"]

        Rsr = (a1 * Pr**a2) + ((1 - a1) * Pr**a3)
        Rs = Solution_GOR_Pb_ValkoMcCain(Psep, Tsep, API, Rsp) * Rsr

    elif P_array >= Pb_Velarde(API, SGsep, Tres, Psep, Tsep, Rsp):
        Rs = 0

    else:
        Rs = "Missing Data"

    if Units == 1:

        Rs = Rs

    else:
        Rs = Rs / (1.78107607 * 10**-1)

    return Rs


# print ("Rs =",Solution_GOR_Velarde2 (SGsep, API, Tres, P, Psep, Tsep, Rsp))


def Co_above_Pb_Spivey1(API, SGsep, P, Pres, Tres, Psep, Tsep, Rsp, Units=1):
    """The Spivey1 correlation calculates the Oil Compressibility from Pb to
    a higher pressure of interest, used for estimating certain fluid
    properties,the required input data is: Stock tank oil gravity [API],
    Separator gas Specific Gravity, Pressure [psia], Reservoir Pressure [
    psia] and Reservoir Temperature [F]. Also requires the Pb from the
    Velarde correlation and Rsb from the ValkoMcCain correlation

        If Units=1 use:
            Pressure in psia
            Temperature in F
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in Scf/STB

        If Units=2 use:
            Pressure in MPa
            Temperature in C
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in m3/m3"""

    if Units == 1:

        API_array = np.array(API)
        SGsep_array = np.array(SGsep)
        P_array = np.array(P)
        Tres_array = np.array(Tres)

    else:
        Tres_array = np.array(Tres)
        Tres_array = (Tres_array * 9 / 5) + 32
        API_array = np.array(API)
        SGsep_array = np.array(SGsep)
        P_array = np.array(P)
        P_array = P_array / (1.45037738 * 10**2)

    if API_array > 0 and SGsep_array > 0 and Tres_array > 0 and P_array > 0:

        Pr = (P_array -
              14.7) / (Pb_Velarde(API, SGsep, Tres, Psep, Tsep, Rsp) - 14.7)

        lnAPI = np.log(API_array)
        lnSGsep = np.log(SGsep_array)
        lnPb = np.log(Pb_Velarde(API, SGsep, Tres, Psep, Tsep, Rsp))
        lnPr = np.log(Pr)
        lnRsb = np.log(Solution_GOR_Pb_ValkoMcCain(Psep, Tsep, API, Rsp))
        lnTres = np.log(Tres_array)

        #         VARn            C0n                C1n                C2n
        Zntable = {
            "VAR1": lnAPI,
            "C01": 3.011,
            "C11": -2.6254,
            "C21": 0.497,
            "VAR2": lnSGsep,
            "C02": -0.0835,
            "C12": -0.259,
            "C22": 0.382,
            "VAR3": lnPb,
            "C03": 3.51,
            "C13": -0.0289,
            "C23": -0.0584,
            "VAR4": lnPr,
            "C04": 0.327,
            "C14": -0.608,
            "C24": 0.0911,
            "VAR5": lnRsb,
            "C05": -1.918,
            "C15": -0.642,
            "C25": 0.154,
            "VAR6": lnTres,
            "C06": 2.52,
            "C16": -2.73,
            "C26": 0.429
        }

        z1 = Zntable["C01"] + (Zntable["C11"] * Zntable["VAR1"]) + (
            Zntable["C21"] * Zntable["VAR1"]**2)
        z2 = Zntable["C02"] + (Zntable["C12"] * Zntable["VAR2"]) + (
            Zntable["C22"] * Zntable["VAR2"]**2)
        z3 = Zntable["C03"] + (Zntable["C13"] * Zntable["VAR3"]) + (
            Zntable["C23"] * Zntable["VAR3"]**2)
        z4 = Zntable["C04"] + (Zntable["C14"] * Zntable["VAR4"]) + (
            Zntable["C24"] * Zntable["VAR4"]**2)
        z5 = Zntable["C05"] + (Zntable["C15"] * Zntable["VAR5"]) + (
            Zntable["C25"] * Zntable["VAR5"]**2)
        z6 = Zntable["C06"] + (Zntable["C16"] * Zntable["VAR6"]) + (
            Zntable["C26"] * Zntable["VAR6"]**2)

        sumZ = z1 + z2 + z3 + z4 + z5 + z6

        Cofb = np.exp(2.434 + (0.475 * sumZ) + (0.048 * sumZ**2) -
                      np.log(10**6))

    else:
        Cofb = "Missing Data"

    if Units == 1:
        Cofb = Cofb

    else:
        Cofb = Cofb / (1.45037738 * 10**2)

    return Cofb


# print ("Cofb =",Co_above_Pb_Spivey1 (API, SGsep, P, Pres, Tres, Psep,
# Tsep, Rsp))


def Co_above_Pb_Spivey2(Pres, P, API, SGsep, Tres, Psep, Tsep, Rsp, Units=1):
    """The Spivey2 correlation calculates the Oil Compressibility from the
    initial reservoir pressure to a lower pressure but higher than Pb,
    used for material balance, the required imput data is: Pressure [psia],
    Reservoir Pressure [psia]. Also requires the Pb from the Velarde
    correlation and Cofb and Cofbi from the Spivey1 correlation

        If Units=1 use:
            Pressure in psia
            Temperature in F
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in Scf/STB

        If Units=2 use:
            Pressure in MPa
            Temperature in C
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in m3/m3"""

    if Units == 1:

        Pres_array = np.array(Pres)
        P_array = np.array(P)
        API_array = np.array(API)
        SGsep_array = np.array(SGsep)
        Tres_array = np.array(Tres)

    else:
        Tres_array = np.array(Tres)
        Tres_array = (Tres_array * 9 / 5) + 32
        API_array = np.array(API)
        SGsep_array = np.array(SGsep)
        P_array = np.array(P)
        P_array = P_array / (1.45037738 * 10**2)
        Pres_array = np.array(Pres)
        Pres_array = Pres_array / (1.45037738 * 10**2)

    if Pres_array > 0 and P_array > 0:

        Pri = (Pres_array -
               14.7) / (Pb_Velarde(API, SGsep, Tres, Psep, Tsep, Rsp) - 14.7)

        lnAPI = np.log(API_array)
        lnSGsep = np.log(SGsep_array)
        lnPb = np.log(Pb_Velarde(API, SGsep, Tres, Psep, Tsep, Rsp))
        lnPri = np.log(Pri)
        lnRsb = np.log(Solution_GOR_Pb_ValkoMcCain(Psep, Tsep, API, Rsp))
        lnTres = np.log(Tres_array)
        #      VARn             C0n               C1n            C2n
        Zntable = {
            "VAR1": lnAPI,
            "C01": 3.011,
            "C11": -2.6254,
            "C21": 0.497,
            "VAR2": lnSGsep,
            "C02": -0.0835,
            "C12": -0.259,
            "C22": 0.382,
            "VAR3": lnPb,
            "C03": 3.51,
            "C13": -0.0289,
            "C23": -0.0584,
            "VAR4": lnPri,
            "C04": 0.327,
            "C14": -0.608,
            "C24": 0.0911,
            "VAR5": lnRsb,
            "C05": -1.918,
            "C15": -0.642,
            "C25": 0.154,
            "VAR6": lnTres,
            "C06": 2.52,
            "C16": -2.73,
            "C26": 0.429
        }

        z1 = Zntable["C01"] + (Zntable["C11"] * Zntable["VAR1"]) + (
            Zntable["C21"] * Zntable["VAR1"]**2)
        z2 = Zntable["C02"] + (Zntable["C12"] * Zntable["VAR2"]) + (
            Zntable["C22"] * Zntable["VAR2"]**2)
        z3 = Zntable["C03"] + (Zntable["C13"] * Zntable["VAR3"]) + (
            Zntable["C23"] * Zntable["VAR3"]**2)
        z4 = Zntable["C04"] + (Zntable["C14"] * Zntable["VAR4"]) + (
            Zntable["C24"] * Zntable["VAR4"]**2)
        z5 = Zntable["C05"] + (Zntable["C15"] * Zntable["VAR5"]) + (
            Zntable["C25"] * Zntable["VAR5"]**2)
        z6 = Zntable["C06"] + (Zntable["C16"] * Zntable["VAR6"]) + (
            Zntable["C26"] * Zntable["VAR6"]**2)

        sumZ = z1 + z2 + z3 + z4 + z5 + z6

        Cofbi = np.exp(2.434 + (0.475 * sumZ) + (0.048 * sumZ**2) -
                       np.log(10**6))

        Cofi = (
            ((Pb_Velarde(API, SGsep, Tres, Psep, Tsep, Rsp) - P_array) *
             Co_above_Pb_Spivey1(API, SGsep, P, Pres, Tres, Psep, Tsep, Rsp)) -
            ((Pb_Velarde(API, SGsep, Tres, Psep, Tsep, Rsp) - Pres_array) *
             Cofbi)) / (Pres_array - P_array)

    else:
        Cofi = "Missing Data"

    if Units == 1:
        Cofi = Cofi

    else:
        Cofi = Cofi / (1.45037728 * 10**2)

    return Cofi


# print ("Cofi =",Co_above_Pb_Spivey2 (Pres, P, API, SGsep, Tres, Psep,
# Tsep, Rsp))


def Co_above_Pb_Spivey3(P, API, SGsep, Tres, Psep, Tsep, Rsp, Pres, Units=1):
    """The Spivey3 correlation calculates the Oil Compressibility determined
    with the slope of a tangent line to the isothermal density-pressure
    curve at the pressure of interest, used for single phase pressure
    transient analysis, the required imput data is: Pressure [psia]. Also
    requires the Pb from the Velarde correlation, Cofb and Z constant values
    from the Spivey1 correlation

        If Units=1 use:
            Pressure in psia
            Temperature in F
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in Scf/STB

        If Units=2 use:
            Pressure in MPa
            Temperature in C
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in m3/m3"""

    if Units == 1:

        P_array = np.array(P)
        API_array = np.array(API)
        SGsep_array = np.array(SGsep)
        Tres_array = np.array(Tres)

    else:
        Tres_array = np.array(Tres)
        Tres_array = (Tres_array * 9 / 5) + 32
        API_array = np.array(API)
        SGsep_array = np.array(SGsep)
        P_array = np.array(P)
        P_array = P_array / (1.45037738 * 10**2)

    if P_array > 0:

        Pr = (P_array -
              14.7) / (Pb_Velarde(API, SGsep, Tres, Psep, Tsep, Rsp) - 14.7)

        dZdP = (-0.608 + (0.1822 * np.log(P_array / Pb_Velarde(
            API, SGsep, Tres, Psep, Tsep, Rsp)))) / P_array

        lnAPI = np.log(API_array)
        lnSGsep = np.log(SGsep_array)
        lnPb = np.log(Pb_Velarde(API, SGsep, Tres, Psep, Tsep, Rsp))
        lnPr = np.log(Pr)
        lnRsb = np.log(Solution_GOR_Pb_ValkoMcCain(Psep, Tsep, API, Rsp))
        lnTres = np.log(Tres_array)

        #         VARn            C0n                C1n                C2n
        Zntable = {
            "VAR1": lnAPI,
            "C01": 3.011,
            "C11": -2.6254,
            "C21": 0.497,
            "VAR2": lnSGsep,
            "C02": -0.0835,
            "C12": -0.259,
            "C22": 0.382,
            "VAR3": lnPb,
            "C03": 3.51,
            "C13": -0.0289,
            "C23": -0.0584,
            "VAR4": lnPr,
            "C04": 0.327,
            "C14": -0.608,
            "C24": 0.0911,
            "VAR5": lnRsb,
            "C05": -1.918,
            "C15": -0.642,
            "C25": 0.154,
            "VAR6": lnTres,
            "C06": 2.52,
            "C16": -2.73,
            "C26": 0.429
        }

        z1 = Zntable["C01"] + (Zntable["C11"] * Zntable["VAR1"]) + (
            Zntable["C21"] * Zntable["VAR1"]**2)
        z2 = Zntable["C02"] + (Zntable["C12"] * Zntable["VAR2"]) + (
            Zntable["C22"] * Zntable["VAR2"]**2)
        z3 = Zntable["C03"] + (Zntable["C13"] * Zntable["VAR3"]) + (
            Zntable["C23"] * Zntable["VAR3"]**2)
        z4 = Zntable["C04"] + (Zntable["C14"] * Zntable["VAR4"]) + (
            Zntable["C24"] * Zntable["VAR4"]**2)
        z5 = Zntable["C05"] + (Zntable["C15"] * Zntable["VAR5"]) + (
            Zntable["C25"] * Zntable["VAR5"]**2)
        z6 = Zntable["C06"] + (Zntable["C16"] * Zntable["VAR6"]) + (
            Zntable["C26"] * Zntable["VAR6"]**2)

        sumZ = z1 + z2 + z3 + z4 + z5 + z6

        dCofbdP = Co_above_Pb_Spivey1(API, SGsep, P, Pres, Tres, Psep, Tsep,
                                      Rsp) * ((0.475 + (0.096 * sumZ)) * dZdP)
        Co = Co_above_Pb_Spivey1(
            API, SGsep, P, Pres, Tres, Psep, Tsep, Rsp) + (
                (P_array - Pb_Velarde(API, SGsep, Tres, Psep, Tsep, Rsp)) *
                dCofbdP)

    else:
        Co = "Missing Data"

    if Units == 1:
        Co = Co

    else:
        Co = Co / (1.45037728 * 10**2)

    return Co


# print ("Co =",Co_above_Pb_Spivey3 (P, API, SGsep, Tres, Psep, Tsep, Rsp,
# Pres))

# ----CONTINUE AFTER FVF CORRELATIONS----


def Den_Pb(SGsep, Tres, Psep, Tsep, API, Rsp, Units=1):
    """The Den_Pb function calculates the reservoir oil Densities at
    reservoir pressures equal to bubble point pressure, the required imput
    data is: Separator gas Specific Gravity and Reservoir Temperature [F].
    Also requires the Rsb from ValkoMcCain correlation, SG and SGst from
    ValkoMcCain2 correlation and Pb from the Velarde correlation

        If Units=1 use:
            Pressure in psia
            Temperature in F
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in Scf/STB

        If Units=2 use:
            Pressure in MPa
            Temperature in C
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in m3/m3"""

    if Units == 1:

        SGsep_array = np.array(SGsep)
        Tres_array = np.array(Tres)

    else:
        Tres_array = np.array(Tres)
        Tres_array = (Tres_array * 9 / 5) + 32
        SGsep_array = np.array(SGsep)

    Den_pseudo1 = 52.8 - (0.01 *
                          (Solution_GOR_Pb_ValkoMcCain(Psep, Tsep, API, Rsp)))

    if SGsep_array > 0 and Tres_array > 0:

        antable = {
            "a0": -49.8930,
            "a1": 85.0149,
            "a2": -3.70373,
            "a3": 0.0479818,
            "a4": 2.98914,
            "a5": -0.0356888
        }

        den_ap = antable["a0"] + antable["a1"] * SGsep_array + antable[
            "a2"] * SGsep_array * Den_pseudo1 + antable[
                "a3"] * SGsep_array * Den_pseudo1**2 + antable[
                    "a4"] * Den_pseudo1 + antable["a5"] * Den_pseudo1**2

        Den_pseudo2 = (
            (Solution_GOR_Pb_ValkoMcCain(Psep, Tsep, API, Rsp) *
             Spec_grav_ValkoMcCain2(Psep, Tsep, API, Rsp, SGsep)) +
            (4600 * Spec_grav_st_ValkoMcCain2(Psep, Tsep, API, Rsp, SGsep))
        ) / (73.71 +
             ((Solution_GOR_Pb_ValkoMcCain(Psep, Tsep, API, Rsp) *
               Spec_grav_ValkoMcCain2(Psep, Tsep, API, Rsp, SGsep)) / den_ap))

        while ((abs(Den_pseudo1 - Den_pseudo2)) / Den_pseudo2) > 0.1:

            Den_pseudo1 = Den_pseudo2

            den_ap = antable["a0"] + antable["a1"] * SGsep_array + antable[
                "a2"] * SGsep_array * Den_pseudo1 + antable[
                    "a3"] * SGsep_array * Den_pseudo1**2 + antable[
                        "a4"] * Den_pseudo1 + antable["a5"] * Den_pseudo1**2

            Den_pseudo2 = (
                (Solution_GOR_Pb_ValkoMcCain(Psep, Tsep, API, Rsp) *
                 Spec_grav_ValkoMcCain2(Psep, Tsep, API, Rsp, SGsep)) +
                (4600 * Spec_grav_st_ValkoMcCain2(Psep, Tsep, API, Rsp, SGsep))
            ) / (73.71 +
                 ((Solution_GOR_Pb_ValkoMcCain(Psep, Tsep, API, Rsp) *
                   Spec_grav_ValkoMcCain2(Psep, Tsep, API, Rsp, SGsep)) /
                  den_ap))

            if ((abs(Den_pseudo1 - Den_pseudo2)) / Den_pseudo2) < 0.1:
                break

        den_adjP = (
            (0.167 + 16.181 * (10**(-0.0425 * Den_pseudo2))) *
            (Pb_Velarde(API, SGsep, Tres, Psep, Tsep, Rsp) / 1000)) - 0.01 * (
                (0.299 + 263 * (10**(-0.0603 * Den_pseudo2))) *
                (Pb_Velarde(API, SGsep, Tres, Psep, Tsep, Rsp) / 1000)**2)

        den_fake = den_adjP + Den_pseudo2

        den_adjT = ((0.00302 + (1.505 * den_fake**-0.951)) * (
            (Tres_array - 60)**0.938)) - ((0.0216 - 0.0233 *
                                           (10**(-0.0161 * den_fake))) *
                                          ((Tres_array - 60)**0.475))

        den_Pb = den_fake - den_adjT

    if Units == 1:
        den_Pb = den_Pb

    else:
        den_Pb = den_Pb / (1.60184634 * 10**-2)

    return den_Pb


# print ("den_Pb =",Den_Pb (SGsep, Tres, Psep, Tsep, API, Rsp))


def Den_underPb(SGsep, Tres, P, API, Psep, Tsep, Rsp, Units=1):
    """The Den_underPb function calculates the reservoir oil Densities at
    reservoir pressures less than bubble point pressure, the required imput
    data is: Separator gas Specific Gravity, Reservoir Temperature [F] and
    Pressure [psia]. Also requires the Rs from Velarde2 correlation and SG
    and SGst from ValkoMcCain2 correlation

        If Units=1 use:
            Pressure in psia
            Temperature in F
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in Scf/STB

        If Units=2 use:
            Pressure in MPa
            Temperature in C
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in m3/m3"""

    if Units == 1:
        SGsep_array = np.array(SGsep)
        Tres_array = np.array(Tres)
        P_array = np.array(P)

    else:
        Tres_array = np.array(Tres)
        Tres_array = (Tres_array * 9 / 5) + 32
        SGsep_array = np.array(SGsep)
        P_array = np.array(P)
        P_array = P_array / (1.45037738 * 10**2)

    Den_pseudo1 = 52.8 - (
        0.01 * (Solution_GOR_Velarde2(SGsep, API, Tres, P, Psep, Tsep, Rsp)))

    if SGsep_array > 0 and Tres_array > 0:

        antable = {
            "a0": -49.8930,
            "a1": 85.0149,
            "a2": -3.70373,
            "a3": 0.0479818,
            "a4": 2.98914,
            "a5": -0.0356888
        }

        den_ap = antable["a0"] + antable["a1"] * SGsep_array + antable[
            "a2"] * SGsep_array * Den_pseudo1 + antable[
                "a3"] * SGsep_array * Den_pseudo1**2 + antable[
                    "a4"] * Den_pseudo1 + antable["a5"] * Den_pseudo1**2

        Den_pseudo2 = (
            (Solution_GOR_Velarde2(SGsep, API, Tres, P, Psep, Tsep, Rsp) *
             Spec_grav_ValkoMcCain2(Psep, Tsep, API, Rsp, SGsep)) +
            (4600 * Spec_grav_st_ValkoMcCain2(Psep, Tsep, API, Rsp, SGsep))
        ) / (73.71 +
             ((Solution_GOR_Velarde2(SGsep, API, Tres, P, Psep, Tsep, Rsp) *
               Spec_grav_ValkoMcCain2(Psep, Tsep, API, Rsp, SGsep)) / den_ap))

        while ((abs(Den_pseudo1 - Den_pseudo2)) / Den_pseudo2) * 100 > 10:

            Den_pseudo1 = Den_pseudo2

            den_ap = antable["a0"] + antable["a1"] * SGsep_array + antable[
                "a2"] * SGsep_array * Den_pseudo1 + antable[
                    "a3"] * SGsep_array * Den_pseudo1**2 + antable[
                        "a4"] * Den_pseudo1 + antable["a5"] * Den_pseudo1**2

            Den_pseudo2 = (
                (Solution_GOR_Velarde2(SGsep, API, Tres, P, Psep, Tsep, Rsp) *
                 Spec_grav_ValkoMcCain2(Psep, Tsep, API, Rsp, SGsep)) +
                (4600 * Spec_grav_st_ValkoMcCain2(Psep, Tsep, API, Rsp, SGsep))
            ) / (73.71 + (
                (Solution_GOR_Velarde2(SGsep, API, Tres, P, Psep, Tsep, Rsp) *
                 Spec_grav_ValkoMcCain2(Psep, Tsep, API, Rsp, SGsep)) /
                den_ap))

            if ((abs(Den_pseudo1 - Den_pseudo2)) / Den_pseudo2) < 0.1:
                break

        den_adjP = ((0.167 + 16.181 * (10**(-0.0425 * Den_pseudo2))) *
                    (P_array / 1000)) - 0.01 * (
                        (0.299 + 263 *
                         (10**(-0.0603 * Den_pseudo2))) * (P_array / 1000)**2)

        den_fake = den_adjP + Den_pseudo2

        den_adjT = ((0.00302 + (1.505 * den_fake**-0.951)) * (
            (Tres_array - 60)**0.938)) - ((0.0216 - 0.0233 *
                                           (10**(-0.0161 * den_fake))) *
                                          ((Tres_array - 60)**0.475))

        den_underPb = den_fake - den_adjT

    if Units == 1:
        den_underPb = den_underPb

    else:
        den_underPb = den_underPb / (1.60184634 * 10**-2)

    return den_underPb


# print ("den_underPb =",Den_underPb (SGsep, Tres, P, API, Psep, Tsep, Rsp))


def Den_abovePb(P, SGsep, Tres, Psep, Tsep, API, Rsp, Pres, Units=1):
    """The Den_abovePb function calculates the reservoir oil Densities at
    reservoir pressures greater than bubble point pressure, the required
    imput data is: Pressure [psia]. Also requires the Den_Pb function,
    Cofb from Spivey1 correlation and Pb from Velarde correlation

        If Units=1 use:
            Pressure in psia
            Temperature in F
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in Scf/STB

        If Units=2 use:
            Pressure in MPa
            Temperature in C
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in m3/m3"""

    if Units == 1:

        P_array = np.array(P)

    else:

        P_array = np.array(P)
        P_array = P_array / (1.45037738 * 10**2)

    den_abovePb = Den_Pb(SGsep, Tres, Psep, Tsep, API, Rsp) * np.exp(
        Co_above_Pb_Spivey1(API, SGsep, P, Pres, Tres, Psep, Tsep, Rsp) *
        (P_array - Pb_Velarde(API, SGsep, Tres, Psep, Tsep, Rsp)))

    if Units == 1:
        den_abovePb = den_abovePb

    else:
        den_abovePb = den_abovePb / (1.60184634 * 10**-2)

    return den_abovePb


# print("den_abovePb =",Den_abovePb(P, SGsep, Tres, Psep, Tsep, API, Rsp,
# Pres))


def FVF_Pb(Den_sto, Psep, Tsep, API, Rsp, SGsep, Tres, Units=1):
    """The FVF_Pb function calculates the Formation Volume Factor at
    reservoir pressure equal to bubblepoint pressure, the required imput
    data is: Density of stock-tank oil [lb/cu ft]. Also requires the Rsb
    from ValkoMcCain correlation, SG from ValkoMcCain2 correlation, Den_Pb
    function

        If Units=1 use:
            Pressure in psia
            Temperature in F
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in Scf/STB
            Density @ Stock-Tank in lb/ft3

        If Units=2 use:
            Pressure in MPa
            Temperature in C
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in m3/m3
            Density @ Stock-Tank in gr/cm3"""

    if Units == 1:

        Den_sto_array = np.array(Den_sto)

    else:
        Den_sto_array = np.array(Den_sto)
        Den_sto_array = Den_sto_array / (1.60184634 * 10**-2)

    Bob = (Den_sto_array +
           (0.01357 * Solution_GOR_Pb_ValkoMcCain(Psep, Tsep, API, Rsp) *
            Spec_grav_st_ValkoMcCain2(Psep, Tsep, API, Rsp, SGsep))) / Den_Pb(
                SGsep, Tres, Psep, Tsep, API, Rsp)

    return Bob


# print ("Bob =",FVF_Pb (Den_sto, Psep, Tsep, API, Rsp, SGsep, Tres))


def FVF_underPb(Den_sto, Psep, Tsep, API, Rsp, SGsep, Tres, P, Units=1):
    """The FVF_underPb function calculates the Formation Volume Factor at
    reservoir pressure less than bubblepoint pressure, the required imput
    data is: Density of stock-tank oil [lb/cu ft]. Also requires the Rs from
    Velarde correlation, SG from ValkoMcCain2 correlation and Den_underPb
    function

        If Units=1 use:
            Pressure in psia
            Temperature in F
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in Scf/STB
            Density @ Stock-Tank in lb/ft3

        If Units=2 use:
            Pressure in MPa
            Temperature in C
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in m3/m3
            Density @ Stock-Tank in gr/cm3"""

    if Units == 1:

        Den_sto_array = np.array(Den_sto)

    else:
        Den_sto_array = np.array(Den_sto)
        Den_sto_array = Den_sto_array / (1.60184634 * 10**-2)

    Bo_underPb = (Den_sto_array + (
        0.01357 * Solution_GOR_Velarde2(SGsep, API, Tres, P, Psep, Tsep, Rsp) *
        Spec_grav_st_ValkoMcCain2(Psep, Tsep, API, Rsp, SGsep))) / Den_underPb(
            SGsep, Tres, P, API, Psep, Tsep, Rsp)

    return Bo_underPb


# print ("Bo_underPb =",FVF_underPb(Den_sto, Psep, Tsep, API, Rsp, SGsep,
# Tres, P))


def FVF_abovePb(P, Den_sto, Psep, Tsep, API, Rsp, SGsep, Tres, Pres, Units=1):
    """The FVF_underPb function calculates the Formation Volume Factor at
    reservoir pressure greater than bubblepoint pressure, the required imput
    data is: Pressure [psia]. Also requires the Bob from FVF_Pb function,
    Cofb from Spivey1 correlation and Pb from Velarde correlation

        If Units=1 use:
            Pressure in psia
            Temperature in F
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in Scf/STB
            Density @ Stock-Tank in lb/ft3

        If Units=2 use:
            Pressure in MPa
            Temperature in C
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in m3/m3
            Density @ Stock-Tank in gr/cm3"""

    if Units == 1:

        P_array = np.array(P)

    else:
        P_array = np.array(P)
        P_array = P_array / (1.45037738 * 10**2)

    Bo_abovePb = FVF_Pb(Den_sto, Psep, Tsep, API, Rsp, SGsep, Tres) * np.exp(
        Co_above_Pb_Spivey1(API, SGsep, P, Pres, Tres, Psep, Tsep, Rsp) *
        (Pb_Velarde(API, SGsep, Tres, Psep, Tsep, Rsp) - P_array))

    return Bo_abovePb


# print ("Bo_abovePb =",FVF_abovePb (P, Den_sto, Psep, Tsep, API, Rsp,
# SGsep, Tres, Pres))

# ----CONTINUE OF VISCOSITY CORRELATIONS----


def Viscosity_Pb(API, Tres, Psep, Tsep, Rsp, Units=1):
    """The Viscosity_Pb function calculates the Viscosity at reservoir
    pressures equal to bubblepoint pressure, the required imput data is:
    Stock tank oil gravity [API] and Reservoir Temperature [F]. Also
    requires the Rsb from ValkoMcCain correlation

        If Units=1 use:
            Pressure in psia
            Temperature in F
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in Scf/STB

        If Units=2 use:
            Pressure in MPa
            Temperature in C
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in m3/m3"""

    if Units == 1:

        API_array = np.array(API)
        Tres_array = np.array(Tres)

    else:
        API_array = np.array(API)
        Tres_array = np.array(Tres)
        Tres_array = (Tres_array * 9 / 5) + 32

    C = (10**(3.0324 - 0.02023 * API_array)) * (Tres_array**-1.163)

    uoD = (10**C) - 1

    A = 10.715 * (Solution_GOR_Pb_ValkoMcCain(Psep, Tsep, API, Rsp) +
                  100)**-0.515
    B = 5.44 * (Solution_GOR_Pb_ValkoMcCain(Psep, Tsep, API, Rsp) +
                150)**-0.338

    uob = A * uoD**B

    if Units == 1:
        uob = uob

    else:
        uob = uob / (0.001)

    return uob


# print("uob =",Viscosity_Pb (API, Tres, Psep, Tsep, Rsp))


def Viscosity_underPb(SGsep, API, Tres, P, Psep, Tsep, Rsp, Units=1):
    """The Viscosity_underPb function calculates the Viscosity at reservoir
    pressures under bubblepoint pressure. Requires the Rs from Velarde2
    correlation and uoR from Viscosity_Pb function

        If Units=1 use:
            Pressure in psia
            Temperature in F
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in Scf/STB

        If Units=2 use:
            Pressure in MPa
            Temperature in C
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in m3/m3"""

    if Units == 1:

        API_array = np.array(API)
        Tres_array = np.array(Tres)

    else:
        API_array = np.array(API)
        Tres_array = np.array(Tres)
        Tres_array = (Tres_array * 9 / 5) + 32

    A = 10.715 * (Solution_GOR_Velarde2(SGsep, API, Tres, P, Psep, Tsep, Rsp) +
                  100)**-0.515
    B = 5.44 * (Solution_GOR_Velarde2(SGsep, API, Tres, P, Psep, Tsep, Rsp) +
                150)**-0.338

    C = (10**(3.0324 - 0.02023 * API_array)) * (Tres_array**-1.163)

    uoD = (10**C) - 1

    uoR = A * uoD**B

    if Units == 1:
        uoR = uoR

    else:
        uoR = uoR / (0.001)

    return uoR


# print("uoR_underPb =",Viscosity_underPb (SGsep, API, Tres, P, Psep, Tsep,
# Rsp))


def Viscosity_abovePb(P, API, Tres, Psep, Tsep, Rsp, SGsep, Units=1):
    """The Viscosity_abovePb function calculates the Viscosity at reservoir
    pressures above bubblepoint pressure, the required imput data is:
    Pressure [psia]. Also requires the uob from Viscosity_Pb function and Pb
    from Velarde correlation

        If Units=1 use:
            Pressure in psia
            Temperature in F
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in Scf/STB

        If Units=2 use:
            Pressure in MPa
            Temperature in C
            Stock-Tank oil gravity in API
            Solution Gas Oil Ratio @ Pb in m3/m3"""

    if Units == 1:

        P_array = np.array(P)

    else:
        P_array = np.array(P)
        P_array = P_array / (1.45037738 * 10**2)

    A = -1.0146 + (
        1.3322 * np.log10(Viscosity_Pb(API, Tres, Psep, Tsep, Rsp))) - (
            0.4876 *
            (np.log10(Viscosity_Pb(API, Tres, Psep, Tsep, Rsp)))**2) - (
                1.15036 *
                (np.log10(Viscosity_Pb(API, Tres, Psep, Tsep, Rsp)))**3)

    uoR = Viscosity_Pb(API, Tres, Psep, Tsep, Rsp) + (
        (1.3449 * 10**-3) *
        (P_array - Pb_Velarde(API, SGsep, Tres, Psep, Tsep, Rsp)) * 10**A)

    if Units == 1:
        uoR = uoR

    else:
        uoR = uoR / (0.001)

    return uoR


# print("uoR_abovePb =",Viscosity_abovePb (P, API, Tres, Psep, Tsep, Rsp,
# SGsep))
"----------------------------Water Correlations-----------------------------"


def RS_bw(P, T, salinity, unit=1):
    """This function calculate the Rs of methane dissolved brine The user
    must select UNITS 1 to English System and 2 to International System

    IF UNITS 1 is selected introduce:
        | Pressure (P) in PSI
        | Temperature (T) in ºF
        | Salinity in Weight Fraction
        | Rs returns in scf/STB

    IF UNITS 2 is selected introduce:
        | Pressure (P) in MPa
        | Temperature (T) in ºC
        | Salinity in Weight Fraction
        | Rs returns in cm3/cm3"""

    if unit == 1:
        P_array = np.array(P)
        P_array = P_array * (6.89475729 * 10**-3.0)
        T_array = np.array(T)
        T_array = (T_array - 32) * 5 / 9
        m = np.array(salinity)
        m = 1000 * (m / 1000000) / (58.44 * (1 - (m / 1000000)))

    else:
        P_array = np.array(P)
        T_array = np.array(T)
        m = np.array(salinity)
        m = 1000 * (m / 1000000) / (58.44 * (1 - (m / 1000000)))

    # Step1: Calculate the vapor pressure of pure water

    T_array = T_array + 273

    # Coefficients Table 4_9
    a1Po = -7.85951783
    a2Po = 1.84408259
    a3Po = -11.7866497
    a4Po = 22.6807411
    a5Po = -15.9618719
    a6Po = 1.80122502

    # Ec 4.13
    vad = 1 - T_array / 674.096  # Adimensional T EC.4.14
    # Ec 4.14
    Po = 22.064 * (np.exp((674.096 / T_array) *
                          (a1Po * vad + a2Po * vad**1.5 + a3Po * vad**3 +
                           a4Po * vad**3.5 + a5Po * vad**4 + a6Po * vad**7.5)))

    # Step2: Calculate the solubility of methane in pure water at (P,T)

    T_array = T_array - 273

    # Coefficients Table 4_10 EC 4.1
    a1A_T = 0
    a2A_T = -0.004462
    a3A_T = -0.06763
    a4A_T = 0
    a5A_T = 0

    A_T = (a1A_T * (T_array / 100)**2 + a2A_T *
           (T_array / 100) + a3A_T) / (a4A_T * (T_array / 100)**2 + a5A_T *
                                       (T_array / 100) + 1)

    a1B_T = -0.03602
    a2B_T = 0.18917
    a3B_T = 0.97242
    a4B_T = 0
    a5B_T = 0

    B_T = (a1B_T * (T_array / 100)**2 + a2B_T *
           (T_array / 100) + a3B_T) / (a4B_T * (T_array / 100)**2 + a5B_T *
                                       (T_array / 100) + 1)

    a1C_T = 0.6855
    a2C_T = -3.1992
    a3C_T = -3.7968
    a4C_T = 0.07711
    a5C_T = 0.2229

    C_T = (a1C_T * (T_array / 100)**2 + a2C_T *
           (T_array / 100) + a3C_T) / (a4C_T * (T_array / 100)**2 + a5C_T *
                                       (T_array / 100) + 1)

    # Ec 4.13 (Solubility of methane in pure water)
    mCH4pw = np.exp(A_T * (np.log(P_array - Po))**2 +
                    B_T * np.log(P_array - Po) + C_T)  # Ec 4.15

    T_array = T_array + 273

    # Step3: Calculate the solubility of methane in brine water at (P,T)

    # Coefficients Table 4_11
    lambC1 = -0.80898
    lambC2 = 1.0827 * 10**(-3)
    lambC3 = 183.85
    lambC6 = 3.924 * 10**(-4)
    lambC10 = -1.97 * 10**(-6)

    # Ec 4.16
    lambCH4 = (lambC1 + lambC2 * T_array + lambC3 / T_array + lambC6 * P_array
               + lambC10 * P_array**2)  # EC4.16

    # Ec 4.18
    mCH4bw = mCH4pw * np.exp(-2 * lambCH4 * m - (-3.89 * 10**(-3)) * m**2)

    Z = 0.98  # TODO input for user?

    # Ec 4.34
    VMCH4gsc = Z * 8.314467 * (15.5 + 273) / 0.001

    # Step4: Calculate the formation volumen factor

    # Ec 4.36

    if unit == 1:
        Rs = (mCH4bw * VMCH4gsc) / ((1000 + m * 58.44) * 0.99681786)
        Rs = Rs * 5.61458333

    else:
        Rs = (mCH4bw * VMCH4gsc) / ((1000 + m * 58.44) * 0.99681786)

    return Rs


def Rs_McCain(P, T, salinity):
    """The Rs_McCain function calculates the solubilty of gas on water using
    the correlations of McCain

        IF UNITS 1 is selected introduce:
            | Pressure (P) in PSI
            | Temperature (T) in ºF
            | Salinity in Weight Fraction
            | Rs returns in scf/STB

        IF UNITS 2 is selected introduce:
            | Pressure (P) in MPa
            | Temperature (T) in ºC
            | Salinity in Weight Fraction
            | Rs returns in cm3/cm3"""
    """if unit == 1:
        #P_array = np.array(P)"""
    P_array = P * (6.89475729 * 10**-3.0)
    # T_array = np.array(T)
    T_array = (T - 32) * 5 / 9
    # m = np.array(salinity)
    # m = 1000 * (m / 1000000) / (58.44 * (1 - (m / 1000000)))
    """else:
        #P_array = np.array(P)
        P_array = P
        #T_array = np.array(T)
        T_array = T
        #m = np.array(salinity)
        #m = 1000 * (m / 1000000) / (58.44 * (1 - (m / 1000000)))"""

    A = 8.15839 - (6.12265e-2 * T_array) + (1.191633e-4 * T_array**2) - (
        2.1654e-7 * T_array**3)
    B = (1.011021e-2) - (7.44241e-5 * T_array) + (3.05553e-7 * T_array**2) - (
        2.94883e-10 * T_array**3)
    C = -1e-7 * (9.02505 - 0.13023 * T_array + (8.53425e-4 * T_array**2) -
                 (2.34122e-6 * T_array**3) - (2.37049e-9 * T_array**4))

    Rswp = A + B * P_array + C * P_array

    Rsw = 0.0840655 * salinity * T_array * Rswp

    return Rsw


def Bo_bw(P, T, salinity, unit=1):
    """This function calculate the Bo of methane dissolved brine
    The user must select UNITS 1 to Field units and 2 to International System

    IF UNITS 1 is selected introduce:
        | Pressure (P) in PSI
        | Temperature (T)2 in ºF
        | Salinity in ppm
        | Bo returns in bbl/STB

    IF UNITS 2 is selected introduce:
        Pressure (P) in MPa
        Temperature (T) in ºC
        Salinity in ppm
        Bo returns in cm3/cm3"""

    # Step1: Calculate the derivatives of Duan

    if unit == 1:
        P_array = np.array(P)
        P_array = P_array * (6.89475729 * 10**-3.0)
        T_array = np.array(T)
        T_array = (T_array - 32) * 5 / 9
        m = np.array(salinity)
        m = 1000 * (m / 1000000) / (58.44 * (1 - (m / 1000000)))

    else:
        P_array = np.array(P)
        T_array = np.array(T)
        m = np.array(salinity)
        m = 1000 * (m / 1000000) / (58.44 * (1 - (m / 1000000)))

    T_array = T_array + 273
    # Coefficients Table 4_11
    UCH4c6 = 7.6985890 * 10**(-2)
    UCH4c7 = -5.0253331 * 10**(-5)
    UCH4c8 = -30.092013
    UCH4c9 = 4.8468502 * 10**(3)

    CH4c6 = 3.92 * 10**(-4)
    CH4c10 = -1.97 * 10**(-6)

    # Ec 4.19
    dch4dp = UCH4c6 + UCH4c7 * T_array + UCH4c8 / T_array + UCH4c9 / (
        T_array)**2
    # Ec 4.20
    dlamdp = CH4c6 + 2 * CH4c10 * P_array
    # Ec 4.22 Partial molar volume of methane in brine
    VMCH4B = 8.314467 * T_array * (dch4dp + 2 * m * dlamdp + m**2 * 0)

    # Step2: Calculate the density of methane-free brine water

    T_array = T_array - 273

    # Coefficients Table 4_6 and Ec 4.1
    a1_T_70 = -0.127213
    a2_T_70 = 0.645486
    a3_T_70 = 1.03265
    a4_T_70 = -0.070291
    a5_T_70 = 0.639589

    a_T_70 = (a1_T_70 * (T_array / 100)**2 + a2_T_70 *
              (T_array / 100) + a3_T_70) / (a4_T_70 *
                                            (T_array / 100)**2 + a5_T_70 *
                                            (T_array / 100) + 1)

    a1_EW = 4.221
    a2_EW = -3.478
    a3_EW = 6.221
    a4_EW = 0.5182
    a5_EW = -0.4405

    a_T_EW = (a1_EW * (T_array / 100)**2 + a2_EW *
              (T_array / 100) + a3_EW) / (a4_EW * (T_array / 100)**2 + a5_EW *
                                          (T_array / 100) + 1)

    a1_FW = -11.403
    a2_FW = 29.932
    a3_FW = 27.952
    a4_FW = 0.20684
    a5_FW = 0.3768

    a_T_FW = (a1_FW * (T_array / 100)**2 + a2_FW *
              (T_array / 100) + a3_FW) / (a4_FW * (T_array / 100)**2 + a5_FW *
                                          (T_array / 100) + 1)

    # Coefficients Table 4_7 and Ec 4.1

    a1_DM2 = -1.1149 * 10**(-4)
    a2_DM2 = 1.7105 * 10**(-4)
    a3_DM2 = -4.3766 * 10**(-4)
    a4_DM2 = 0
    a5_DM2 = 0

    a_T_DM2 = (a1_DM2 * (T_array / 100)**2 + a2_DM2 *
               (T_array / 100) + a3_DM2) / (a4_DM2 *
                                            (T_array / 100)**2 + a5_DM2 *
                                            (T_array / 100) + 1)

    a1_DM3_2 = -8.878 * 10**(-4)
    a2_DM3_2 = -1.388 * 10**(-4)
    a3_DM3_2 = -2.96318 * 10**(-3)
    a4_DM3_2 = 0
    a5_DM3_2 = 0.51103

    a_T_DM3_2 = (a1_DM3_2 * (T_array / 100)**2 + a2_DM3_2 *
                 (T_array / 100) + a3_DM3_2) / (a4_DM3_2 *
                                                (T_array / 100)**2 + a5_DM3_2 *
                                                (T_array / 100) + 1)

    a1_DM1 = 2.1466 * 10**(-3)
    a2_DM1 = 1.2427 * 10**(-2)
    a3_DM1 = 4.2648 * 10**(-2)
    a4_DM1 = -8.1009 * 10**(-2)
    a5_DM1 = 0.525417

    a_T_DM1 = (a1_DM1 * (T_array / 100)**2 + a2_DM1 *
               (T_array / 100) + a3_DM1) / (a4_DM1 *
                                            (T_array / 100)**2 + a5_DM1 *
                                            (T_array / 100) + 1)

    a1_DM1_2 = 2.366 * 10**(-4)
    a2_DM1_2 = -3.636 * 10**(-4)
    a3_DM1_2 = -2.278 * 10**(-4)
    a4_DM1_2 = 0
    a5_DM1_2 = 0

    a_T_DM1_2 = (a1_DM1_2 * (T_array / 100)**2 + a2_DM1_2 *
                 (T_array / 100) + a3_DM1_2) / (a4_DM1_2 *
                                                (T_array / 100)**2 + a5_DM1_2 *
                                                (T_array / 100) + 1)

    # Ec 4.6 Density of brine water at (70MPa and T)
    dbw_T_70 = a_T_70 + a_T_DM2 * 1 * m**2 + a_T_DM3_2 * 1 * m**(
        3 / 2) + a_T_DM1 * 1 * m + a_T_DM1_2 * 1 * m**(1 / 2)

    # Coefficients Table 4_8 Ec 4.1

    a1_EM = 0
    a2_EM = 0
    a3_EM = 0.1249
    a4_EM = 0
    a5_EM = 0

    a_T_EM = (a1_EM * (T_array / 100)**2 + a2_EM *
              (T_array / 100) + a3_EM) / (a4_EM * (T_array / 100)**2 + a5_EM *
                                          (T_array / 100) + 1)

    a1_FM3_2 = -0.617
    a2_FM3_2 = -0.747
    a3_FM3_2 = -0.4339
    a4_FM3_2 = 0
    a5_FM3_2 = 10.26

    a_T_FM3_2 = (a1_FM3_2 * (T_array / 100)**2 + a2_FM3_2 *
                 (T_array / 100) + a3_FM3_2) / (a4_FM3_2 *
                                                (T_array / 100)**2 + a5_FM3_2 *
                                                (T_array / 100) + 1)

    a1_FM1 = 0
    a2_FM1 = 9.917
    a3_FM1 = 5.1128
    a4_FM1 = 0
    a5_FM1 = 3.892

    a_T_FM1 = (a1_FM1 * (T_array / 100)**2 + a2_FM1 *
               (T_array / 100) + a3_FM1) / (a4_FM1 *
                                            (T_array / 100)**2 + a5_FM1 *
                                            (T_array / 100) + 1)

    a1_FM1_2 = 0.0365
    a2_FM1_2 = -0.0369
    a3_FM1_2 = 0
    a4_FM1_2 = 0
    a5_FM1_2 = 0

    a_T_FM1_2 = (a1_FM1_2 * (T_array / 100)**2 + a2_FM1_2 *
                 (T_array / 100) + a3_FM1_2) / (a4_FM1_2 *
                                                (T_array / 100)**2 + a5_FM1_2 *
                                                (T_array / 100) + 1)

    Eb_T_M = a_T_EW + a_T_EM * m

    FB_T_M = a_T_FW + a_T_FM3_2 * m**(3 /
                                      2) + a_T_FM1 * m + a_T_FM1_2 * m**(1 / 2)

    Ib_T_70_m = (1 / Eb_T_M) * np.log(np.abs(Eb_T_M + FB_T_M))

    Ib_T_P_m = (1 / Eb_T_M) * np.log(np.abs(Eb_T_M * (P_array / 70) + FB_T_M))

    # Density of methane-free brine wate
    denbwnogas = (dbw_T_70 * np.exp((Ib_T_P_m - Ib_T_70_m)))

    # Step3: Calculate the vapor pressure of pure water

    T_array = T_array + 273

    # Coefficients Table 4_9
    a1Po = -7.85951783
    a2Po = 1.84408259
    a3Po = -11.7866497
    a4Po = 22.6807411
    a5Po = -15.9618719
    a6Po = 1.80122502

    # Ec 4.13
    vad = 1 - T_array / 674.096  # Adimensional T EC.4.14
    # Ec 4.14
    Po = 22.064 * (np.exp((674.096 / T_array) *
                          (a1Po * vad + a2Po * vad**1.5 + a3Po * vad**3 +
                           a4Po * vad**3.5 + a5Po * vad**4 + a6Po * vad**7.5)))

    # Step4: Calculate the solubility of methane in pure water at (P,T)

    T_array = T_array - 273

    # Coefficients Table 4_10 EC 4.1
    a1A_T = 0
    a2A_T = -0.004462
    a3A_T = -0.06763
    a4A_T = 0
    a5A_T = 0

    A_T = (a1A_T * (T_array / 100)**2 + a2A_T *
           (T_array / 100) + a3A_T) / (a4A_T * (T_array / 100)**2 + a5A_T *
                                       (T_array / 100) + 1)

    a1B_T = -0.03602
    a2B_T = 0.18917
    a3B_T = 0.97242
    a4B_T = 0
    a5B_T = 0

    B_T = (a1B_T * (T_array / 100)**2 + a2B_T *
           (T_array / 100) + a3B_T) / (a4B_T * (T_array / 100)**2 + a5B_T *
                                       (T_array / 100) + 1)

    a1C_T = 0.6855
    a2C_T = -3.1992
    a3C_T = -3.7968
    a4C_T = 0.07711
    a5C_T = 0.2229

    C_T = (a1C_T * (T_array / 100)**2 + a2C_T *
           (T_array / 100) + a3C_T) / (a4C_T * (T_array / 100)**2 + a5C_T *
                                       (T_array / 100) + 1)

    # Ec 4.13 (Solubility of methane in pure water)
    mCH4pw = np.exp(A_T * (np.log(P_array - Po))**2 +
                    B_T * np.log(P_array - Po) + C_T)  # Ec 4.15

    T_array = T_array + 273

    # Step5: Calculate the solubility of methane in brine water at (P,T)

    # Coefficients Table 4_11
    lambC1 = -0.80898
    lambC2 = 1.0827 * 10**(-3)
    lambC3 = 183.85
    lambC6 = 3.924 * 10**(-4)
    lambC10 = -1.97 * 10**(-6)

    # Ec 4.16
    lambCH4 = (lambC1 + lambC2 * T_array + lambC3 / T_array + lambC6 * P_array
               + lambC10 * P_array**2)  # EC4.16

    # Ec 4.18
    mCH4bw = mCH4pw * np.exp(-2 * lambCH4 * m - (-3.89 * 10**(-3)) * m**2)

    # Ec. 4.23 Specific Volume of methane-free
    Vbw = 1 / (denbwnogas)

    # Step6: Calculate the formation volume factor

    # Ec 4.36

    if unit == 1:
        Bw = ((1000 + m * 58.44) * Vbw + mCH4bw * VMCH4B) / (
            (1000 + m * 58.44) * 0.99681786)
        Bw = Bw * 1

    else:
        Bw = ((1000 + m * 58.44) * Vbw + mCH4bw * VMCH4B) / (
            (1000 + m * 58.44) * 0.99681786)

    return Bw


def comp_bw_nogas(P, T, salinity, unit=1):
    """This function calculate the compressibility of methane-free brine
    The user must select UNITS 1 to Field units and 2 to International System

    IF UNITS 1 is selected introduce:
        | Pressure (P) in PSI
        | Temperature (T) in ºF
        | Salinity in ppm
        | Compressibility returns in psi-1

    IF UNITS 2 is selected introduce:
        | Pressure (P) in MPa
        | Temperature (T) in ºC
        | Salinity in ppm
        | Compressibility returns in MPa-1"""

    if unit == 1:
        P_array = np.array(P)
        P_array = P_array * (6.89475729 * 10**-3.0)
        T_array = np.array(T)
        T_array = (T_array - 32) * 5 / 9
        m = np.array(salinity)
        m = 1000 * (m / 1000000) / (58.44 * (1 - (m / 1000000)))
    else:
        P_array = np.array(P)
        T_array = np.array(T)
        m = np.array(salinity)
        m = 1000 * (m / 1000000) / (58.44 * (1 - (m / 1000000)))

    # Step 1: Calculate the compressibility coefficients

    # Ec 4.1 and coefficients Table 4_6
    a1_EW = 4.221
    a2_EW = -3.478
    a3_EW = 6.221
    a4_EW = 0.5182
    a5_EW = -0.4405

    a_T_EW = (a1_EW * (T_array / 100)**2 + a2_EW *
              (T_array / 100) + a3_EW) / (a4_EW * (T_array / 100)**2 + a5_EW *
                                          (T_array / 100) + 1)

    a1_FW = -11.403
    a2_FW = 29.932
    a3_FW = 27.952
    a4_FW = 0.20684
    a5_FW = 0.3768

    a_T_FW = (a1_FW * (T_array / 100)**2 + a2_FW *
              (T_array / 100) + a3_FW) / (a4_FW * (T_array / 100)**2 + a5_FW *
                                          (T_array / 100) + 1)

    # Ec 4.1 and coefficients Table 4_8

    a1_EM = 0
    a2_EM = 0
    a3_EM = 0.1249
    a4_EM = 0
    a5_EM = 0

    a_T_EM = (a1_EM * (T_array / 100)**2 + a2_EM *
              (T_array / 100) + a3_EM) / (a4_EM * (T_array / 100)**2 + a5_EM *
                                          (T_array / 100) + 1)

    a1_FM3_2 = -0.617
    a2_FM3_2 = -0.747
    a3_FM3_2 = -0.4339
    a4_FM3_2 = 0
    a5_FM3_2 = 10.26

    a_T_FM3_2 = (a1_FM3_2 * (T_array / 100)**2 + a2_FM3_2 *
                 (T_array / 100) + a3_FM3_2) / (a4_FM3_2 *
                                                (T_array / 100)**2 + a5_FM3_2 *
                                                (T_array / 100) + 1)

    a1_FM1 = 0
    a2_FM1 = 9.917
    a3_FM1 = 5.1128
    a4_FM1 = 0
    a5_FM1 = 3.892

    a_T_FM1 = (a1_FM1 * (T_array / 100)**2 + a2_FM1 *
               (T_array / 100) + a3_FM1) / (a4_FM1 *
                                            (T_array / 100)**2 + a5_FM1 *
                                            (T_array / 100) + 1)

    a1_FM1_2 = 0.0365
    a2_FM1_2 = -0.0369
    a3_FM1_2 = 0
    a4_FM1_2 = 0
    a5_FM1_2 = 0

    a_T_FM1_2 = (a1_FM1_2 * (T_array / 100)**2 + a2_FM1_2 *
                 (T_array / 100) + a3_FM1_2) / (a4_FM1_2 *
                                                (T_array / 100)**2 + a5_FM1_2 *
                                                (T_array / 100) + 1)

    # Ec 4.7
    Eb_T_M = a_T_EW + a_T_EM * m
    # Ec 4.8
    FB_T_M = a_T_FW + a_T_FM3_2 * m**(3 /
                                      2) + a_T_FM1 * m + a_T_FM1_2 * m**(1 / 2)

    # Step 2: Calculate the Compressibility of methane-free Brine at (P,T,m)
    # Ec 4.9  (Compressibility of methane-free Brine at (P,T,m))

    if unit == 1:
        temp = 1 / 70 * (1 / (Eb_T_M * (P_array / 70) + FB_T_M))
        temp = temp * 6.89475729 * 10**(-3.0)

    else:
        temp = 1 / 70 * (1 / (Eb_T_M * (P_array / 70) + FB_T_M))

    return temp


API = 35
SGsep = 0.75
Den_sto = 0.85
P = np.array([200, 100, 150, 50])
Pres = 2000
Tres = 80
Psep = 1500
Tsep = 70
Rsp = 100
Units = 1
dos = RS_bw(200, 250, 3000, 1)
# print("solucion:",dos)



# pytank/__init__.py
from pytank.vector_data import ProdVector, PressVector
from pytank.functions.utilities import normalize_date_freq
from pytank.functions.helpers import create_wells, search_wells
from pytank.well import Well
from pytank.fluid_model import OilModel, WaterModel
from pytank.tank import Tank
from pytank.analysis import Analysis

__all__ = [
    'ProdVector',
    'PressVector',
    'normalize_date_freq',
    'create_wells',
    'search_wells',
    'Well',
    'OilModel',
    'WaterModel',
    'Tank',
    'Analysis',
]



"""
aquifer_model.py

This module defines the Fetkovich and CarterTracy Class to calculate
cumulative influx of water.

The logic is structured using classes and method

libraries:
    - pandas
    - math
    - numpy
"""

import pandas as pd
import numpy as np
import math
from pytank.functions.utilities import variable_type
from datetime import datetime
from typing import Optional


# Fetkovich
class Fetkovich:
    """
    To estimate water influx using the Fetkovich's method we need to estimate:
    Wi, Wei and J.
    Parameters
    ----------
    aq_radius : float
        Radius of the aquifer, ft.
    res_radius : float
        Radius of the reservoir, ft.
    aq_thickness : float
        Thickness of the aquifer, ft.
    aq_por : float
        Porosity of the aquifer.
    ct : float
        Total compressibility coefficient, psi^-1.
    theta : float
        Encroachment angle.
    k : float
        Permeability of the aquifer, md.
    water_visc : float
        Viscosity of water, cp.
    boundary_type : str, optional
        Type of aquifer boundary, default is 'no_flow'. Options are 'no_flow',
        'constant_pressure', 'infinite'.
    flow_type : str, optional
        Type of flow, default is 'radial'. Options are 'radial', 'linear'.
    pr : list or numpy array, optional
        Measured reservoir pressure, psi.
    time_step : list or numpy array, optional
        Time step, days.
    width : float, optional
        Width of the linear aquifer, ft. Required only for linear flow.
    length : float, optional
        Length of the linear aquifer, ft. Required only for linear flow.

    Returns
    -------
    pandas.DataFrame
        Containing the cumulative water influx, bbl.
    """

    def __init__(
            self,
            aq_radius: float,
            res_radius: float,
            aq_thickness: float,
            aq_por: float,
            ct: float,
            theta: float,
            k: float,
            water_visc: float,
            boundary_type: str = "no_flow",
            flow_type: str = "radial",
            pr: Optional[list] = None,
            time_step: Optional[list] = None,
            width: float = None,
            length: float = None,
    ):
        """
        Initializes the attributes of the Fetkovich's class.

    Parameters
    ----------
    aq_radius : float
        Radius of the aquifer, ft.
    res_radius : float
        Radius of the reservoir, ft.
    aq_thickness : float
        Thickness of the aquifer, ft.
    aq_por : float
        Porosity of the aquifer.
    ct : float
        Total compressibility coefficient, psi^-1.
    theta : float
        Encroachment angle.
    k : float
        Permeability of the aquifer, md.
    water_visc : float
        Viscosity of water, cp.
    boundary_type : str, optional
        Type of aquifer boundary, default is 'no_flow'. Options are 'no_flow',
         'constant_pressure', 'infinite'.
    flow_type : str, optional
        Type of flow, default is 'radial'. Options are 'radial', 'linear'.
    pr : list or numpy array, optional
        Measured reservoir pressure, psi.
    time_step : list or numpy array, optional
        Time step, days.
    width : float, optional
        Width of the linear aquifer, ft. Required only for linear flow.
    length : float, optional
        Length of the linear aquifer, ft. Required only for linear flow.
        """
        self.aq_radius = aq_radius
        self.res_radius = res_radius
        self.aq_thickness = aq_thickness
        self.aq_por = aq_por
        self.ct = ct
        self.theta = theta
        self.k = k
        self.water_visc = water_visc
        self.boundary_type = boundary_type
        self.flow_type = flow_type
        self.pr = pr
        self.time_step = time_step
        self.width = width
        self.length = length

    def _set_pr_and_time_step(self, pr, time_step):
        """
        Private method to assign value o the pr and time_step properties of
        the aquifer Fetkovich
        :param
        pr: List of Measured reservoir pressure, psi.
        time_step: List of Time lapses, days.
        """

        self.pr = pr
        self.time_step = time_step

        # Check if the time list is in datetime format
        if all(isinstance(t, datetime) for t in time_step):
            # If all elements are already datetime objects, convert -
            # to cumulative days
            self.time_step = [(t - time_step[0]).days for t in time_step]
        elif all(isinstance(t, pd.Timestamp) for t in time_step):
            # If elements are Timestamp objects, convert to datetime and -
            # then to cumulative days
            datetime_list = [pd.to_datetime(t) for t in time_step]
            self.time_step = [(t - datetime_list[0].days
                               for t in datetime_list)]
        elif all(isinstance(t, str) for t in time_step):
            # If all elements are strings, convert to datetime and then -
            # to cumulative days
            datetime_list = [
                datetime.strptime(t, "%Y-%m-%d") for t in time_step
            ]
            self.time_step = [(t - datetime_list[0]).days
                              for t in datetime_list]
        else:
            # Convert dates to cumulative days using variable_types
            self.time_step = variable_type(time_step)

        # Automatically influx of water data upon object creation
        self.we()

    def we(self) -> pd.DataFrame:
        """
        Method to calculate cumulative influx of water and gets the values -
        in a DataFrame.

        :return:
        pd.Dataframe: DataFrame wit the delta WE, Cumulative WE and time
         lapses columns.
        """
        if self.flow_type == "linear" and (self.width is None
                                           or self.length is None):
            raise ValueError("When using linear flow, "
                             "width and length are required arguments")
        # Check if pressure and time step are arrays, list or floats
        pr_array = variable_type(self.pr)
        delta_t = variable_type(self.time_step)

        # if not all(pr_array[i] >= pr_array[i + 1]
        #           for i in range(len(pr_array) - 1)):
        #    raise ValueError("Pressure array must be in descendant order")

        # Check if pressure array is not in descendant order throw an error
        if not all(pr_array > 0):
            raise ValueError("Pressure must be greater than zero")
        if isinstance(pr_array, np.ndarray) and isinstance(
                delta_t, np.ndarray):
            dim_pr = np.size(pr_array)
            dim_time = np.size(delta_t)
            if dim_pr != dim_time:
                raise ValueError("Dimensions of pressure array and time array "
                                 "should be equal,"
                                 "please verify your input")

        # Calculate the initial volume of water in the aquifers (Wi)
        wi = ((math.pi / 5.615) * (self.aq_radius ** 2 - self.res_radius ** 2)
              * self.aq_thickness * self.aq_por)

        # Calculate the maximum possible water influx (Wei)
        f = self.theta / 360
        wei = self.ct * wi * pr_array[0] * f

        # Calculate the aquifers productivity index
        # based on the boundary_type conditions and aquifers geometry (J)
        rd = self.aq_radius / self.res_radius

        j = None

        if self.boundary_type == "no_flow" and self.flow_type == "radial":
            j = (0.00708 * self.k * self.aq_thickness *
                 f) / (self.water_visc * (math.log(rd) - 0.75))
        elif (self.boundary_type == "constant_pressure" and self.flow_type ==
              "radial"):
            j = (0.00708 * self.k * self.aq_thickness * f) / (self.water_visc *
                                                              math.log(rd))
        elif self.boundary_type == "no_flow" and self.flow_type == "linear":
            j = (0.003381 * self.k * self.width *
                 self.aq_thickness) / (self.water_visc * self.length)
        elif (self.boundary_type == "constant_pressure" and self.flow_type ==
              "linear"):
            j = (0.001127 * self.k * self.width *
                 self.aq_thickness) / (self.water_visc * self.length)
        elif self.boundary_type == "infinite" and self.flow_type == "radial":
            a = math.sqrt(
                (0.0142 * self.k * 365) / (f * self.water_visc * self.ct))
            j = (0.00708 * self.k * self.aq_thickness *
                 f) / (self.water_visc * math.log(a / self.res_radius))

        # Calculate the incremental water influx (We)n during the nth -
        # time interval

        # Calculate cumulative water influx
        cum_water_influx = 0
        pr = pr_array[0]

        # Average aquifers pressure after removing We bbl of water from -
        # the aquifers
        pa = pr_array[0]
        elapsed_time = np.empty((1, 0))
        time_steps = np.array(0)
        df_list = []
        for ip in range(len(pr_array)):
            pr_avg = (pr + pr_array[ip]) / 2
            if isinstance(delta_t, np.ndarray):
                diff_pr = np.diff(delta_t)
                time_steps = np.append(time_steps, diff_pr)
                we = ((wei / pr_array[0]) * (1 - math.exp(
                    (-1 * j * pr_array[0] * time_steps[ip]) / wei)) *
                      (pa - pr_avg))
                elapsed_time = delta_t
            else:
                we = ((wei / pr_array[0]) * (1 - math.exp(
                    (-1 * j * pr_array[0] * delta_t) / wei)) * (pa - pr_avg))
                elapsed_time = np.append(elapsed_time, delta_t * ip)
            pr = pr_array[ip]
            cum_water_influx = cum_water_influx + we
            pa = pr_array[0] * (1 - (cum_water_influx / wei))

            # Creation values for each key of the list df_list that will -
            # have a dictionary
            df_list.append({
                "Delta We": we,
                "Cumulative We": cum_water_influx,
                "Elapsed time": elapsed_time[ip],
            })

        # Creation of the dataframe that will be return for users
        df = pd.DataFrame(df_list)
        # df = df.set_index("Elapsed time")
        return df

    def get_we(self) -> pd.Series:
        """
        Method than encapsulated a DataFrame to gets an only column
        of Cumulative We.
        :return:
        pd.Series: A column of values of Cumulative We.
        """
        # Encapsulation of DataFrame using method We.
        we = self.we()
        return we["Cumulative We"]


class CarterTracy:
    """
    Class to calculate the cumulative influx of water through Carter Tracy
    aquifer model.

    Parameters
    ----------
    aq_por : float
        Porosity of the aquifer (decimal).
    ct : float
        Total compressibility, psi^-1.
    res_radius : float
        Radius of the reservoir, ft.
    aq_thickness : float
        Thickness of the aquifer, ft.
    theta : float
        Encroachment angle, degrees.
    aq_perm : float
        Permeability of the aquifer, md.
    water_visc : float
        Viscosity of water, cp.
    pr : list, optional
        List of Measured reservoir pressure, psi.
    time_step : list, optional
        Time lapses, days.

    Returns
    -------
    pandas.DataFrame
        Containing the cumulative water influx, bbl.
    """

    def __init__(
            self,
            aq_por: float,
            ct: float,
            res_radius: float,
            aq_thickness: float,
            theta: float,
            aq_perm: float,
            water_visc: float,
            pr: Optional[list] = None,
            time_step: Optional[list] = None,
    ):
        """
        Initializes the attributes of the CarterTracy class.

        Parameters
        ----------
        aq_por : float
            Porosity of the aquifer (decimal).
        ct : float
            Total compressibility, psi^-1.
        res_radius : float
            Radius of the reservoir, ft.
        aq_thickness : float
            Thickness of the aquifer, ft.
        theta : float
            Encroachment angle, degrees.
        aq_perm : float
            Permeability of the aquifer, md.
        water_visc : float
            Viscosity of water, cp.
        pr : list, optional
            Measured reservoir pressure, psi.
        time_step : list, optional
            Time lapses, days.
        """
        self.aq_por = aq_por
        self.ct = ct
        self.res_radius = res_radius
        self.aq_thickness = aq_thickness
        self.theta = theta
        self.aq_perm = aq_perm
        self.water_visc = water_visc
        self.time = time_step
        self.pr = pr

    def _set_pr_and_time_step(self, pr, time_step):
        """
        Private method to assign value to the pr and time_step
        properties of the aquifer Carter Tracy
        :param
        pr: List of Measured reservoir pressure, psi.
        time_step: List of Time lapses, days.
        """

        self.pr = pr
        self.time_step = time_step
        # Check if the time list is in datetime format
        if all(isinstance(t, datetime) for t in time_step):
            # If all elements are already datetime objects, convert to -
            # cumulative days
            self.time = [(t - time_step[0]).days for t in time_step]
        elif all(isinstance(t, pd.Timestamp) for t in time_step):
            # If elements are Timestamp objects, convert to datetime and then -
            # to cumulative days
            datetime_list = [pd.to_datetime(t) for t in time_step]
            self.time = [(t - datetime_list[0].days for t in datetime_list)]
        elif all(isinstance(t, str) for t in time_step):
            # If all elements are strings, convert to datetime and then -
            # to cumulative days
            datetime_list = [
                datetime.strptime(t, "%Y-%m-%d") for t in time_step
            ]
            self.time = [(t - datetime_list[0]).days for t in datetime_list]
        else:
            # Convert dates to cumulative days using variable_types
            self.time = variable_type(time_step)

        # Automatically influx of water data upon object creation
        self.we()

    def we(self) -> pd.DataFrame:
        """
        Method to calculate cumulative influx of water and gets the values -
        in a DataFrame.

        :return:
        pd.Dataframe: DataFrame wit the delta WE, Cumulative WE and time
        lapses columns.
        """

        # Check if pressure and time are arrays, lists or floats
        pr_array = variable_type(self.pr)
        t_array = variable_type(self.time)

        # Check if pressure array is not in descendant order throw an error
        # if not all(pr_array[i] >= pr_array[i + 1]
        #           for i in range(len(pr_array) - 1)):
        #    raise ValueError("Pressure array must be in descendant order")

        # Check if pressure array is not in descendant order throw an error
        if not all(pr_array > 0):
            raise ValueError("Pressure must be greater than zero")

        # Check if time step and pressure dimensions are equal
        # this can be done if time step is entered as array
        if isinstance(pr_array, np.ndarray) and isinstance(
                t_array, np.ndarray):
            dim_pr = np.size(pr_array)
            dim_time = np.size(t_array)
            if dim_pr != dim_time:
                raise ValueError("Dimensions of pressure array and time array "
                                 "should be equal,"
                                 "please verify your input")
        # Calculate the van Everdingen-Hurst water influx constant
        f = self.theta / 360
        b = 1.119 * self.aq_por * self.ct * (self.res_radius **
                                             2) * self.aq_thickness * f
        # Estimate dimensionless time (tD)
        cte = (0.006328 * self.aq_perm /
               (self.aq_por * self.water_visc * self.ct *
                (self.res_radius ** 2)))
        td = np.where(t_array > 0, t_array * cte, 0)

        # Calculate the total pressure drop (Pi-Pn) as an array, for each  -
        # time step n.
        pr_drop = np.where(pr_array > 0, pr_array[0] - pr_array, 1)

        # Estimate the dimensionless pressure
        pr_d = np.where(
            td > 100,
            0.5 * (np.log(np.maximum(td, 1e-15)) + 0.80907),
            ((370.529 * np.sqrt(td)) + (137.582 * td) +
             (5.69549 * (td ** 1.5))) / (328.834 + (265.488 * np.sqrt(td)) +
                                         (45.2157 * td) + (td ** 1.5)),
        )
        # Estimate the dimensionless pressure derivative
        e = 716.441 + (46.7984 * (td * 0.5)) + (270.038 * td) + (71.0098 *
                                                                 (td * 1.5))
        d = ((1296.86 * (td ** 0.5)) + (1204.73 * td) + (618.618 * (td * 1.5))
             + (538.072 * (td * 2)) + (142.41 * (td ** 2.5)))
        pr_deriv = np.where(td > 100, 1 / (2 * np.maximum(td, 1e-15)),
                            e / np.maximum(d, 1e-15))

        # Calculate the cumulative water influx at any time, ti
        df = {"Cumulative We": [0]}
        we = 0

        for i in np.arange(1, len(td)):
            a1 = td[i] - td[i - 1]
            a2 = b * pr_drop[i]
            a3 = we * pr_deriv[i]
            a4 = pr_d[i]
            a5 = td[i - 1] * pr_deriv[i]
            cum_influx_water = we + (a1 * ((a2 - a3) / (a4 - a5)))
            we = cum_influx_water
            df["Cumulative We"].append(we)

        df["Elapsed time, days"] = t_array

        # Concatenation of the DataFrames in a unique final DataFrame
        df = pd.concat([pd.DataFrame(df)], ignore_index=True)

        return df

    def get_we(self) -> pd.Series:
        """
        Method than encapsulated a DataFrame to gets an only column of
        Cumulative We.

        :return:
        pd.Series: A column of values of Cumulative We
        """
        we = self.we()
        return we["Cumulative We"]
